# coding: utf-8
"""module defining utils for an edf volume"""

from __future__ import annotations

import os
import logging
import mmap

import fabio
import fabio.edfimage
import numpy
from silx.io.url import DataUrl

from tomoscan.esrf.identifier.edfidentifier import EDFVolumeIdentifier
from tomoscan.esrf.volume.singleframebase import (
    VolumeSingleFrameBase,
    TIFFLikeDiskAccessor,
)
from tomoscan.scanbase import TomoScanBase
from tomoscan.utils import docstring

_logger = logging.getLogger(__name__)


__all__ = [
    "EDFVolume",
]


class EDFVolume(TIFFLikeDiskAccessor, VolumeSingleFrameBase):
    """
    Save volume data to single frame edf and metadata to .txt files

    :warning: each file saved under {volume_basename}_{index_zfill6}.edf is considered to be a slice of the volume.
    """

    DEFAULT_DATA_SCHEME = "fabio"

    DEFAULT_DATA_EXTENSION = "edf"

    def __init__(
        self,
        folder: str | None = None,
        volume_basename: str | None = None,
        data: numpy.ndarray | None = None,
        source_scan: TomoScanBase | None = None,
        metadata: dict | None = None,
        data_url: DataUrl | None = None,
        metadata_url: DataUrl | None = None,
        overwrite: bool = False,
        header: dict | None = None,
        start_index=0,
        data_extension=DEFAULT_DATA_EXTENSION,
        metadata_extension=VolumeSingleFrameBase.DEFAULT_METADATA_EXTENSION,
    ) -> None:
        if folder is not None:
            url = DataUrl(
                file_path=str(folder),
                data_path=None,
            )
        else:
            url = None
        TIFFLikeDiskAccessor.__init__(self)
        super().__init__(
            volume_basename=volume_basename,
            url=url,
            data=data,
            source_scan=source_scan,
            metadata=metadata,
            data_url=data_url,
            metadata_url=metadata_url,
            overwrite=overwrite,
            start_index=start_index,
            data_extension=data_extension,
            metadata_extension=metadata_extension,
        )

        self._header = header

    @property
    def header(self) -> dict | None:
        """possible header for the edf files"""
        return self._header

    @docstring(VolumeSingleFrameBase)
    def save_frame(self, frame, file_name, scheme):
        if scheme == "fabio":
            header = self.header or {}
            edf_writer = fabio.edfimage.EdfImage(
                data=frame,
                header=header,
            )
            parent_dir = os.path.dirname(file_name)
            if parent_dir != "":
                os.makedirs(parent_dir, exist_ok=True)
            edf_writer.write(file_name)
        else:
            raise ValueError(f"scheme {scheme} is not handled")

    @docstring(VolumeSingleFrameBase)
    def load_frame(self, file_name, scheme):
        if scheme == "fabio":
            return fabio.open(file_name).data
        else:
            raise ValueError(f"scheme {scheme} is not handled")

    @staticmethod
    @docstring(VolumeSingleFrameBase)
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, EDFVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {EDFVolumeIdentifier}"
            )
        return EDFVolume(
            folder=identifier.folder,
            volume_basename=identifier.file_prefix,
        )

    @docstring(VolumeSingleFrameBase)
    def get_identifier(self) -> EDFVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")
        return EDFVolumeIdentifier(
            object=self, folder=self.url.file_path(), file_prefix=self._volume_basename
        )

    @staticmethod
    def example_defined_from_str_identifier() -> str:
        return " ; ".join(
            [
                f"{EDFVolume(folder='/path/to/my/my_folder').get_identifier().to_str()}",
                f"{EDFVolume(folder='/path/to/my/my_folder', volume_basename='mybasename').get_identifier().to_str()} (if mybasename != folder name)",
            ]
        )

    # implementation of TIFFLikeAccessor

    def read_n_lines_in_file(self, file_name, line_indices) -> tuple:
        reader = _EDFImageWithReadMMap()
        reader.read(file_name)
        volume_shape = self.get_volume_shape()
        result = []
        for line in line_indices:
            data = reader.read_mmap(
                file_name,
                coords=(
                    slice(0, volume_shape[1]),
                    slice(line, line + 1),
                ),
            ).ravel()
            if data is None:
                # an error might have occur ('fast_read_roi' only accept un-compressed). In this case try the 'good old' .data directly which is safer
                data = reader.data[:, line]
            result.append(data)
        return tuple(result)

    def read_n_columns_in_file(self, file_name, column_indices) -> tuple:
        image = fabio.open(file_name)
        volume_shape = self.get_volume_shape()
        result = []
        for column in column_indices:
            data = image.fast_read_roi(
                file_name,
                coords=[
                    slice(column, column + 1),
                    slice(0, volume_shape[2]),
                ],
            ).ravel()

            if data is None:
                # an error might have occur ('fast_read_roi' only accept un-compressed). In this case try the 'good old' .data directly which is safer
                data = image.data[column,]
            result.append(data)
        return tuple(result)

    def read_file(self, file_name) -> tuple:
        return fabio.open(file_name).data


class _EDFImageWithReadMMap(fabio.edfimage.EdfImage):

    def read_mmap(self, fname: str, coords: tuple) -> numpy.ndarray:

        if len(coords) == 4:
            slice1 = self.make_slice(coords)  # and fix that FIXME ...
        elif (
            len(coords) == 2
            and isinstance(coords[0], slice)
            and isinstance(coords[1], slice)
        ):
            slice1 = coords
        else:
            _logger.warning(
                "readROI: Unable to understand Region Of Interest: got %s", coords
            )
            return
        with open(fname, "rb") as fin:
            fin.seek(0)
            count = self.shape[0] * self.shape[1]
            offset = self._frames[self.currentframe].start
            nbytes = self.bpp * count
            with mmap.mmap(
                fin.fileno(), offset + nbytes, access=mmap.ACCESS_READ
            ) as mem:
                buf = numpy.frombuffer(
                    mem, offset=offset, count=count, dtype=self.bytecode
                )
                buf.shape = self.shape
                ret = buf[slice1].copy()
                del buf
        if self.swap_needed():
            ret.byteswap(True)
        return ret



# coding: utf-8

import os
from copy import deepcopy

import h5py
import numpy.random
import pytest
from silx.io.url import DataUrl

from tomoscan.scanbase import ReducedFramesInfos, Source, SourceType, TomoScanBase
from tomoscan.tests.utils import NXtomoMockContext


def test_flat_field_correction(tmp_path):
    # set up
    data_dir = tmp_path / "data"
    data_dir.mkdir()

    scan = TomoScanBase(None, None)
    scan.set_reduced_darks(
        {
            0: numpy.random.random(100).reshape((10, 10)),
        }
    )

    scan.set_reduced_flats(
        {
            1: numpy.random.random(100).reshape((10, 10)),
            12: numpy.random.random(100).reshape((10, 10)),
            21: numpy.random.random(100).reshape((10, 10)),
        }
    )

    data_urls = {}
    projections = {}
    file_path = os.path.join(data_dir, "data_file.h5")

    for i in range(-2, 30):
        projections[i] = numpy.random.random(100).reshape((10, 10))
        data_path = "/".join(("data", str(i)))
        data_urls[i] = DataUrl(file_path=file_path, data_path=data_path, scheme="silx")
        with h5py.File(file_path, mode="a") as h5s:
            h5s[data_path] = projections[i]

    scan.projections = projections

    # test get_flats_weights
    flat_weights = scan._get_flats_weights()
    assert isinstance(flat_weights, dict)
    assert len(flat_weights) == 32
    assert flat_weights.keys() == scan.projections.keys()
    assert flat_weights[-2] == {1: 1.0}
    assert flat_weights[0] == {1: 1.0}
    assert flat_weights[1] == {1: 1.0}
    assert flat_weights[12] == {12: 1.0}
    assert flat_weights[21] == {21: 1.0}
    assert flat_weights[24] == {21: 1.0}

    def assertAlmostEqual(ddict1, ddict2):
        assert ddict1.keys() == ddict2.keys()
        for key in ddict1.keys():
            assert numpy.isclose(ddict1[key], ddict2[key])

    print("flat_weights[2]", flat_weights[2])
    assertAlmostEqual(flat_weights[2], {1: 10.0 / 11.0, 12: 1.0 / 11.0})
    assertAlmostEqual(flat_weights[10], {1: 2.0 / 11.0, 12: 9.0 / 11.0})
    assertAlmostEqual(flat_weights[18], {12: 3.0 / 9.0, 21: 6.0 / 9.0})

    # test flat_field_data_url
    """ensure the flat_field is computed. Simple processing test when
    provided data is a DataUrl"""
    projections_keys = [key for key in scan.projections.keys()]
    projections_urls = [scan.projections[key] for key in projections_keys]
    scan.flat_field_correction(projections_urls, projections_keys)

    # test flat_field_data_numpy_array
    """ensure the flat_field is computed. Simple processing test when
    provided data is a numpy array"""
    scan.projections = data_urls
    projections_keys = [key for key in scan.projections.keys()]
    projections_urls = [scan.projections[key] for key in projections_keys]
    scan.flat_field_correction(projections_urls, projections_keys)


def test_Source_API():
    """Test Source API"""
    source = Source(name="my source", type=SourceType.SYNCHROTRON_X_RAY_SOURCE)
    source.name = "toto"
    with pytest.raises(TypeError):
        source.name = 12
    assert isinstance(source.name, str)
    source.type = SourceType.FREE_ELECTRON_LASER
    assert isinstance(source.type, SourceType)
    source.type = None
    str(source)


def test_TomoScanBase_API():
    """Test TomoScanBase API"""
    with pytest.raises(NotImplementedError):
        TomoScanBase.is_tomoscan_dir("")
    with pytest.raises(NotImplementedError):
        TomoScanBase(scan="", type_="undefined").is_abort()
    scan = TomoScanBase(scan="", type_="undefined")
    scan.source
    scan.flats = {1: numpy.random.random(100 * 100).reshape(100, 100)}
    assert len(scan.flats) == 1

    scan.darks = {0: numpy.random.random(100 * 100).reshape(100, 100)}
    assert len(scan.darks) == 1

    scan.alignment_projections = {
        2: numpy.random.random(100 * 100).reshape(100, 100),
        3: numpy.random.random(100 * 100).reshape(100, 100),
    }
    assert len(scan.alignment_projections) == 2

    for prop in (
        "dark_n",
        "tomo_n",
        "flat_n",
        "pixel_size",
        "instrument_name",
        "dim_1",
        "dim_2",
        "scan_range",
        "ff_interval",
        "energy",
        "intensity_monitor",
        "field_of_view",
        "x_translation",
        "y_translation",
        "z_translation",
        "sequence_name",
        "sample_name",
        "group_size",
        "x_rotation_axis_pixel_position",
    ):
        with pytest.raises(NotImplementedError):
            getattr(scan, prop)

    assert isinstance(scan.to_dict(), dict)

    for fct in (
        "update",
        "get_proj_angle_url",
        "get_projections_intensity_monitor",
        "get_flat_expected_location",
        "get_dark_expected_location",
        "get_projection_expected_location",
        "get_energy_expected_location",
        "get_distance_expected_location",
        "get_pixel_size_expected_location",
    ):
        with pytest.raises(NotImplementedError):
            getattr(scan, fct)()


def test_save_load_reduced_darks(tmpdir):
    with NXtomoMockContext(
        scan_path=os.path.join(tmpdir, "test_save_load_reduced_darks"),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        with pytest.raises(TypeError):
            scan.save_reduced_darks(
                darks=None,
                output_urls=(scan.REDUCED_DARKS_DATAURLS,),
            )

        with pytest.raises(TypeError):
            scan.save_reduced_darks(
                darks={
                    0: numpy.ones((10, 10)),
                },
                output_urls=None,
            )

        scan.path = None
        with pytest.raises(ValueError):
            scan.save_reduced_darks(
                darks={
                    0: numpy.ones((10, 10)),
                },
                output_urls=(scan.REDUCED_DARKS_DATAURLS,),
            )

        with pytest.raises(ValueError):
            scan.save_reduced_darks(
                darks={
                    0: numpy.ones((10, 10)),
                },
                output_urls=(scan.REDUCED_DARKS_DATAURLS,),
            )


def test_ReducedFramesInfo():
    """
    test ReducedFramesInfos class
    """
    infos = ReducedFramesInfos()
    assert infos.to_dict() == {}
    infos.count_time = numpy.array([12.3, 13.0])
    assert infos.count_time == [12.3, 13.0]
    infos.machine_electric_current = [23.5, 56.9]
    assert infos.machine_electric_current == [23.5, 56.9]
    my_dict = deepcopy(infos.to_dict())
    assert my_dict == {
        ReducedFramesInfos.COUNT_TIME_KEY: [12.3, 13.0],
        ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY: [23.5, 56.9],
    }

    infos.clear()
    assert infos.to_dict() == {}
    new_infos = ReducedFramesInfos()
    new_infos.load_from_dict(my_dict)
    assert new_infos.to_dict() == my_dict

    with pytest.raises(TypeError):
        new_infos.count_time = 12
    with pytest.raises(TypeError):
        new_infos.machine_electric_current = 12

    str(new_infos)



# coding: utf-8

import os
import tempfile

import pytest

from tomoscan.esrf.scan.mock import MockNXtomo
from tomoscan.esrf.volume.edfvolume import EDFVolume
from tomoscan.series import (
    Series,
    check_series_is_consistent_frm_sample_name,
    sequences_to_series_from_sample_name,
    series_is_complete_from_group_size,
)


@pytest.mark.parametrize("use_identifiers", [True, False])
def test_series_scan(use_identifiers):
    """simple test of a serie"""
    with tempfile.TemporaryDirectory() as dir_:
        series_1 = Series(use_identifiers=use_identifiers)
        assert isinstance(series_1.name, str)
        series_2 = Series("test", use_identifiers=use_identifiers)
        assert series_2.name == "test"
        assert len(series_2) == 0
        scan1 = MockNXtomo(dir_, n_proj=2).scan
        scan2 = MockNXtomo(dir_, n_proj=2).scan
        series_3 = Series("test", [scan1, scan2], use_identifiers=use_identifiers)
        assert series_3.name == "test"
        assert len(series_3) == 2

        with pytest.raises(TypeError):
            series_1.append("toto")

        assert scan1 not in series_1
        series_1.append(scan1)
        assert len(series_1) == 1
        assert scan1 in series_1
        series_1.append(scan1)
        series_1.remove(scan1)
        series_1.name = "toto"
        with pytest.raises(TypeError):
            series_1.name = 12
        with pytest.raises(TypeError):
            series_1.remove(12)
        series_1.append(scan2)
        series_1.append(scan1)
        assert len(series_1) == 3
        series_1.remove(scan1)
        assert len(series_1) == 2
        series_1 == Series("toto", (scan1, scan2), use_identifiers=use_identifiers)
        assert scan1 in series_1
        assert scan2 in series_1

        identifiers_list = series_1.to_dict_of_str()
        assert type(identifiers_list["objects"]) is list
        assert len(identifiers_list["objects"]) == 2
        for id_str in identifiers_list["objects"]:
            assert isinstance(id_str, str)
        assert series_1 != 12


@pytest.mark.parametrize("use_identifiers", [True, False])
def test_series_volume(use_identifiers):
    volume_1 = EDFVolume(folder="test")
    volume_2 = EDFVolume()
    volume_3 = EDFVolume(folder="test2")
    volume_4 = EDFVolume()

    series_1 = Series("Volume serie", [volume_1, volume_2])
    assert volume_1 in series_1
    assert volume_2 in series_1
    assert volume_3 not in series_1
    assert volume_4 not in series_1
    series_1.remove(volume_2)
    series_1.append(volume_3)

    identifiers_list = series_1.to_dict_of_str()
    assert type(identifiers_list["objects"]) is list
    assert len(identifiers_list["objects"]) == 2
    for id_str in identifiers_list["objects"]:
        assert isinstance(id_str, str)

    series_2 = Series.from_dict_of_str(series_1.to_dict_of_str())
    assert len(series_2) == 2
    with pytest.raises(TypeError):
        Series.from_dict_of_str({"name": "toto", "objects": (12, 13)})


def test_series_utils():
    """test utils function from Series"""
    with tempfile.TemporaryDirectory() as tmp_path:
        dir_1 = os.path.join(tmp_path, "scan1")
        dir_2 = os.path.join(tmp_path, "scan2")
        dir_3 = os.path.join(tmp_path, "scan3")
        for dir_folder in (dir_1, dir_2, dir_3):
            os.makedirs(dir_folder)
        scan_s1_1 = MockNXtomo(dir_1, n_proj=2, sample_name="toto").scan
        scan_s1_2 = MockNXtomo(dir_2, n_proj=2, sample_name="toto").scan
        scan_s2_2 = MockNXtomo(dir_3, n_proj=2, sample_name="titi").scan

        found_series = sequences_to_series_from_sample_name(
            (scan_s1_1, scan_s1_2, scan_s2_2)
        )
        assert len(found_series) == 2
        with pytest.raises(TypeError):
            sequences_to_series_from_sample_name([12])
        for serie in found_series:
            check_series_is_consistent_frm_sample_name(serie)

        with pytest.raises(ValueError):
            check_series_is_consistent_frm_sample_name(
                Series("test", [scan_s1_1, scan_s2_2])
            )

        dir_4 = os.path.join(tmp_path, "scan4")
        dir_5 = os.path.join(tmp_path, "scan5")
        scan_z_series_1 = MockNXtomo(
            dir_4, n_proj=2, sample_name="z-series", group_size=2
        ).scan
        scan_z_series_2 = MockNXtomo(
            dir_5, n_proj=2, sample_name="z-series", group_size=2
        ).scan
        assert not series_is_complete_from_group_size(
            [
                scan_z_series_1,
            ]
        )
        assert series_is_complete_from_group_size([scan_z_series_1, scan_z_series_2])

        dir_6 = os.path.join(tmp_path, "scan6")
        scan_z_series_3 = MockNXtomo(
            dir_6, n_proj=2, sample_name="z-series", group_size=2
        ).scan
        assert series_is_complete_from_group_size(
            [scan_z_series_1, scan_z_series_2, scan_z_series_3]
        )

        with pytest.raises(TypeError):
            series_is_complete_from_group_size([1, 2])



# coding: utf-8

import os

from tomoscan.esrf.identifier.folderidentifier import (
    BaseFolderAndfilePrefixIdentifierMixIn,
)
from tomoscan.identifier import VolumeIdentifier
from tomoscan.utils import docstring

__all__ = ["TIFFVolumeIdentifier", "MultiTiffVolumeIdentifier"]


class TIFFVolumeIdentifier(BaseFolderAndfilePrefixIdentifierMixIn, VolumeIdentifier):
    """Identifier specific to (single frame) tiff volume"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, tomo_type=VolumeIdentifier.TOMO_TYPE)

    @property
    @docstring(VolumeIdentifier)
    def scheme(self) -> str:
        return "tiff"

    @staticmethod
    def from_str(identifier):
        from tomoscan.esrf.volume.tiffvolume import TIFFVolume

        return (
            BaseFolderAndfilePrefixIdentifierMixIn._from_str_to_single_frame_identifier(
                identifier=identifier,
                SingleFrameIdentifierClass=TIFFVolumeIdentifier,
                ObjClass=TIFFVolume,
            )
        )


class MultiTiffVolumeIdentifier(VolumeIdentifier):
    def __init__(self, object, tiff_file):
        super().__init__(object)
        self._file_path = os.path.abspath(os.path.abspath(tiff_file))

    @docstring(VolumeIdentifier)
    def short_description(self) -> str:
        return f"{self.scheme}:{self.tomo_type}:{os.path.basename(self._file_path)}"

    @property
    def file_path(self):
        return self._file_path

    @property
    @docstring(VolumeIdentifier)
    def scheme(self) -> str:
        return "tiff3d"

    def __str__(self):
        return f"{self.scheme}:{self.tomo_type}:{self._file_path}"

    def __eq__(self, other):
        if isinstance(other, MultiTiffVolumeIdentifier):
            return (
                self.tomo_type == other.tomo_type
                and self._file_path == other._file_path
            )
        else:
            return super().__eq__(other)

    def __hash__(self):
        return hash(self._file_path)

    @staticmethod
    def from_str(identifier):
        identifier_no_scheme = identifier.split(":")[-1]
        # TODO: check tomo_type ?
        tiff_file = identifier_no_scheme
        from tomoscan.esrf.volume.tiffvolume import TIFFVolume

        return MultiTiffVolumeIdentifier(object=TIFFVolume, tiff_file=tiff_file)



# coding: utf-8

import os

from tomoscan.identifier import VolumeIdentifier
from tomoscan.utils import docstring

__all__ = [
    "RawVolumeIdentifier",
]


class RawVolumeIdentifier(VolumeIdentifier):
    """Identifier for the .vol volume"""

    def __init__(self, object, file_path):
        super().__init__(object)
        self._file_path = os.path.abspath(os.path.abspath(file_path))

    @docstring(VolumeIdentifier)
    def short_description(self) -> str:
        return f"{self.scheme}:{self.tomo_type}:{os.path.basename(self._file_path)}"

    @property
    def file_path(self):
        return self._file_path

    @property
    @docstring(VolumeIdentifier)
    def scheme(self) -> str:
        return "raw"

    def __str__(self):
        return f"{self.scheme}:{self.tomo_type}:{self._file_path}"

    def __eq__(self, other):
        if isinstance(other, RawVolumeIdentifier):
            return (
                self.tomo_type == other.tomo_type
                and self._file_path == other._file_path
            )
        else:
            return False

    def __hash__(self):
        return hash(self._file_path)

    @staticmethod
    def from_str(identifier):
        identifier_no_scheme = identifier.split(":")[-1]
        vol_file = identifier_no_scheme
        from tomoscan.esrf.volume.rawvolume import RawVolume

        return RawVolumeIdentifier(object=RawVolume, file_path=vol_file)



# coding: utf-8
"""Contains the Factory class and dedicated functions"""

from __future__ import annotations

import os
from urllib.parse import urlparse

from tomoscan.esrf.identifier.jp2kidentifier import JP2KVolumeIdentifier
from tomoscan.esrf.identifier.rawidentifier import RawVolumeIdentifier
from tomoscan.esrf.identifier.tiffidentifier import (
    MultiTiffVolumeIdentifier,
    TIFFVolumeIdentifier,
)
from tomoscan.esrf.identifier.url_utils import split_path
from tomoscan.esrf.volume.edfvolume import EDFVolume
from tomoscan.esrf.volume.hdf5volume import HDF5Volume
from tomoscan.esrf.volume.jp2kvolume import JP2KVolume
from tomoscan.esrf.volume.rawvolume import RawVolume
from tomoscan.esrf.volume.tiffvolume import MultiTIFFVolume, TIFFVolume
from tomoscan.identifier import BaseIdentifier, ScanIdentifier, VolumeIdentifier
from tomoscan.tomoobject import TomoObject

from . import identifier as _identifier_mod
from .esrf.identifier.edfidentifier import EDFTomoScanIdentifier, EDFVolumeIdentifier
from .esrf.identifier.hdf5Identifier import NXtomoScanIdentifier, HDF5VolumeIdentifier
from .esrf.scan.edfscan import EDFTomoScan
from .esrf.scan.nxtomoscan import NXtomoScan
from .scanbase import TomoScanBase
from nxtomo.application.nxtomo import NXtomo as _NXtomo

__all__ = [
    "Factory",
]


class Factory:
    """
    Factory any TomoObject
    """

    @staticmethod
    def create_tomo_object_from_identifier(
        identifier: str | ScanIdentifier,
    ) -> TomoObject:
        """
        Create an instance of TomoScanBase from his identifier if possible

        :param identifier: identifier of the TomoScanBase
        :raises: TypeError if identifier is not a str
        :raises: ValueError if identifier cannot be converted back to an instance of TomoScanBase
        """
        if not isinstance(identifier, (str, BaseIdentifier)):
            raise TypeError(
                f"identifier is expected to be a str or an instance of {BaseIdentifier} not {type(identifier)}. {type(identifier)} provided"
            )

        # step 1: convert identifier to an instance of BaseIdentifier if necessary
        if isinstance(identifier, str):
            info = urlparse(identifier)
            paths = split_path(info.path)
            scheme = info.scheme
            if len(paths) == 1:
                # insure backward compatibility. Originally (until 0.8) there was only one type which was scan
                tomo_type = ScanIdentifier.TOMO_TYPE
            elif len(paths) == 2:
                tomo_type, _ = paths
            else:
                raise ValueError("Failed to parse path string:", info.path)

            if tomo_type == _identifier_mod.VolumeIdentifier.TOMO_TYPE:
                if scheme == "edf":
                    identifier = EDFVolumeIdentifier.from_str(identifier=identifier)
                elif scheme == "hdf5":
                    identifier = HDF5VolumeIdentifier.from_str(identifier=identifier)
                elif scheme == "tiff":
                    identifier = TIFFVolumeIdentifier.from_str(identifier=identifier)
                elif scheme == "tiff3d":
                    identifier = MultiTiffVolumeIdentifier.from_str(
                        identifier=identifier
                    )
                elif scheme == "jp2k":
                    identifier = JP2KVolumeIdentifier.from_str(identifier=identifier)
                elif scheme == "raw":
                    identifier = RawVolumeIdentifier.from_str(identifier=identifier)
                else:
                    raise ValueError(f"Scheme {scheme} is not recognized")

            elif tomo_type == _identifier_mod.ScanIdentifier.TOMO_TYPE:
                # otherwise consider this is a scan. Insure backward compatibility
                if scheme == "edf":
                    identifier = EDFTomoScanIdentifier.from_str(identifier=identifier)
                elif scheme == "hdf5":
                    identifier = NXtomoScanIdentifier.from_str(identifier=identifier)
                else:
                    raise ValueError(f"Scheme {scheme} not recognized")
            else:
                raise ValueError(f"{tomo_type} is not an handled tomo type")

        # step 2: convert identifier to a TomoBaseObject
        assert isinstance(identifier, BaseIdentifier)
        scheme = identifier.scheme
        tomo_type = identifier.tomo_type

        if scheme == "edf":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return EDFVolume.from_identifier(identifier=identifier)
            elif tomo_type == ScanIdentifier.TOMO_TYPE:
                return EDFTomoScan.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError()
        elif scheme == "hdf5":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return HDF5Volume.from_identifier(identifier=identifier)
            elif tomo_type == ScanIdentifier.TOMO_TYPE:
                return NXtomoScan.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError()
        elif scheme == "jp2k":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return JP2KVolume.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError
        elif scheme == "tiff":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return TIFFVolume.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError
        elif scheme == "tiff3d":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return MultiTIFFVolume.from_identifier(identifier=identifier)
            else:
                raise NotImplementedError
        elif scheme == "raw":
            if tomo_type == VolumeIdentifier.TOMO_TYPE:
                return RawVolume.from_identifier(identifier=identifier)
        else:
            raise ValueError(f"Scheme {scheme} not recognized")

    @staticmethod
    def create_scan_object(scan_path: str, entry: str | None = None) -> TomoScanBase:
        """

        :param scan_path: path to the scan directory or file
        :return: ScanBase instance fitting the scan folder or scan path
        """
        # remove any final separator (otherwise basename might fail)
        scan_path = scan_path.rstrip(os.path.sep)
        if entry is None and EDFTomoScan.is_tomoscan_dir(scan_path):
            return EDFTomoScan(scan=scan_path)
        elif NXtomoScan.is_tomoscan_dir(scan_path):
            return NXtomoScan(scan=scan_path, entry=entry)
        else:
            raise ValueError(f"{scan_path} is not a valid scan path")

    @staticmethod
    def create_scan_objects(scan_path: str) -> tuple:
        """

        :param scan_path: path to the scan directory or file
        :return: all possible instances of TomoScanBase contained in the given
                 path
        """
        scan_path = scan_path.rstrip(os.path.sep)
        if EDFTomoScan.is_tomoscan_dir(scan_path):
            return (EDFTomoScan(scan=scan_path),)
        elif NXtomoScan.is_tomoscan_dir(scan_path):
            scans = []
            master_file = NXtomoScan.get_master_file(scan_path=scan_path)
            entries = _NXtomo.get_valid_entries(master_file)
            for entry in entries:
                scans.append(NXtomoScan(scan=scan_path, entry=entry, index=None))
            return tuple(scans)

        raise ValueError(f"{scan_path} is not a valid scan path")

    @staticmethod
    def create_scan_object_frm_dict(_dict: dict) -> TomoScanBase:
        """
        Create a TomoScanBase instance from a dictionary. It should contains
        the TomoScanBase._DICT_TYPE_KEY key at least.

        :param _dict: dictionary to be converted
        :return: instance of TomoScanBase
        """
        if TomoScanBase.DICT_TYPE_KEY not in _dict:
            raise ValueError(
                "given dict is not recognized. Cannot find" "",
                TomoScanBase.DICT_TYPE_KEY,
            )
        elif _dict[TomoScanBase.DICT_TYPE_KEY] == EDFTomoScan._TYPE:
            return EDFTomoScan(scan=None).load_from_dict(_dict)
        else:
            raise ValueError(
                f"Scan type: {_dict[TomoScanBase.DICT_TYPE_KEY]} is not managed"
            )

    @staticmethod
    def is_tomoscan_dir(scan_path: str) -> bool:
        """

        :param scan_path: path to the scan directory or file
        :return: True if the given path is a root folder of an acquisition.
        """
        return NXtomoScan.is_tomoscan_dir(scan_path) or EDFTomoScan.is_tomoscan_dir(
            scan_path
        )



from tomoscan.utils.io import deprecated_warning
from .tests import *  # noqa F401,F403

deprecated_warning(
    type_="Module",
    name="tomoscan.test",
    reason="renamed",
    replacement="tomoscan.tests",
    since_version=2.1,
)



# coding: utf-8
"""module defining utils for a tiff volume"""

from __future__ import annotations

import os

import numpy
from silx.io.dictdump import dicttoini
from silx.io.dictdump import load as load_ini
from silx.io.url import DataUrl

from tomoscan.esrf.identifier.tiffidentifier import (
    MultiTiffVolumeIdentifier,
    TIFFVolumeIdentifier,
)
from tomoscan.esrf.volume.singleframebase import (
    VolumeSingleFrameBase,
    TIFFLikeDiskAccessor,
)
from tomoscan.scanbase import TomoScanBase
from tomoscan.utils import docstring, get_subvolume_shape
from tomoscan.volumebase import VolumeBase, SliceTuple

try:
    import tifffile  # noqa #F401 needed for later possible lazy loading
except ImportError:
    has_tifffile = False
else:
    has_tifffile = True
    from tifffile import TiffWriter
    from tifffile import TiffFile

import logging

_logger = logging.getLogger(__name__)


__all__ = ["check_has_tiffle_file", "TIFFVolume", "MultiTIFFVolume"]


def check_has_tiffle_file(handle_mode: str):
    assert handle_mode in ("warning", "raises")

    if not has_tifffile:
        message = "Unable to import `tifffile`. Unable to load or save tiff file. You can use pip to install it"
        if handle_mode == "message":
            _logger.warning(message)
        elif handle_mode == "raises":
            raise ValueError(message)


class TIFFVolume(TIFFLikeDiskAccessor, VolumeSingleFrameBase):
    """
    Save volume data to single frame tiff and metadata to .txt files

    :warning: each file saved under {volume_basename}_{index_zfill6}.tiff is considered to be a slice of the volume.
    """

    DEFAULT_DATA_EXTENSION = "tiff"

    DEFAULT_DATA_SCHEME = "tifffile"

    def __init__(
        self,
        folder: str | None = None,
        volume_basename: str | None = None,
        data: numpy.ndarray | None = None,
        source_scan: TomoScanBase | None = None,
        metadata: dict | None = None,
        data_url: DataUrl | None = None,
        metadata_url: DataUrl | None = None,
        overwrite: bool = False,
        start_index=0,
        data_extension=DEFAULT_DATA_EXTENSION,
        metadata_extension=VolumeSingleFrameBase.DEFAULT_METADATA_EXTENSION,
    ) -> None:
        if folder is not None:
            url = DataUrl(
                file_path=str(folder),
                data_path=None,
            )
        else:
            url = None
        TIFFLikeDiskAccessor.__init__(self)
        super().__init__(
            url=url,
            volume_basename=volume_basename,
            data=data,
            source_scan=source_scan,
            metadata=metadata,
            data_url=data_url,
            metadata_url=metadata_url,
            overwrite=overwrite,
            start_index=start_index,
            data_extension=data_extension,
            metadata_extension=metadata_extension,
        )

        check_has_tiffle_file("warning")

    @docstring(VolumeSingleFrameBase)
    def save_frame(self, frame, file_name, scheme):
        check_has_tiffle_file("raises")
        if scheme == "tifffile":
            tiff_writer = TiffWriter(file_name)
            tiff_writer.write(frame)
        else:
            raise ValueError(f"scheme {scheme} is not handled")

    @docstring(VolumeSingleFrameBase)
    def load_frame(self, file_name, scheme) -> numpy.ndarray:
        check_has_tiffle_file("raises")
        if scheme == "tifffile":
            return tifffile.imread(file_name)
        else:
            raise ValueError(f"scheme {scheme} is not handled")

    def _get_slices_from_disk(
        self, slices: tuple[SliceTuple], url: DataUrl | None = None
    ) -> dict[SliceTuple, numpy.ndarray]:
        """
        read from files a couple of slices along any axis.
        """
        check_has_tiffle_file("raises")
        return super()._get_slices_from_disk(
            slices=slices,
            url=url,
        )

    # identifier section

    @staticmethod
    @docstring(VolumeSingleFrameBase)
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, TIFFVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {TIFFVolumeIdentifier} not {type(identifier)}"
            )
        return TIFFVolume(
            folder=identifier.folder,
            volume_basename=identifier.file_prefix,
        )

    @docstring(VolumeSingleFrameBase)
    def get_identifier(self) -> TIFFVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")
        return TIFFVolumeIdentifier(
            object=self, folder=self.url.file_path(), file_prefix=self._volume_basename
        )

    @staticmethod
    def example_defined_from_str_identifier() -> str:
        return " ; ".join(
            [
                f"{TIFFVolume(folder='/path/to/my/my_folder').get_identifier().to_str()}",
                f"{TIFFVolume(folder='/path/to/my/my_folder', volume_basename='mybasename').get_identifier().to_str()} (if mybasename != folder name)",
            ]
        )

    # TIFFLikeAccessor implementation

    def read_n_lines_in_file(self, file_name, line_indices) -> tuple:
        return tuple(
            [
                tifffile.imread(file_name, selection=(slice(None, None), index))
                for index in line_indices
            ]
        )

    def read_n_columns_in_file(self, file_name, column_indices):
        return tuple(
            [
                tifffile.imread(file_name, selection=(index, slice(None, None)))
                for index in column_indices
            ]
        )

    def read_file(self, file_name) -> tuple:
        return tifffile.imread(file_name)


class MultiTIFFVolume(VolumeBase):
    """
    Save tiff into a single tiff file

    :param file_path: path to the multiframe tiff file
    """

    def __init__(
        self,
        file_path: str | None = None,
        data: numpy.ndarray | None = None,
        source_scan: TomoScanBase | None = None,
        metadata: dict | None = None,
        data_url: DataUrl | None = None,
        metadata_url: DataUrl | None = None,
        overwrite: bool = False,
        append: bool = False,
    ) -> None:
        if file_path is not None:
            url = DataUrl(file_path=file_path)
        else:
            url = None
        super().__init__(
            url, data, source_scan, metadata, data_url, metadata_url, overwrite
        )
        check_has_tiffle_file("warning")
        self.append = append

    @docstring(VolumeBase)
    def deduce_data_and_metadata_urls(self, url: DataUrl | None) -> tuple:
        # convention for tiff multiframe:
        # expect the url to provide a path to a the tiff multiframe file. so data_url will be the same as url
        # and the metadata_url will target a prefix_info.txt file with prefix is the tiff file prefix

        if url is None:
            return None, None
        else:
            if url.data_slice() is not None:
                raise ValueError(f"data_slice is not handled by the {MultiTIFFVolume}")
            file_path = url.file_path()
            if url.data_path() is not None:
                raise ValueError("data_path is not handled")

            scheme = url.scheme() or "tifffile"
            metadata_file = "_".join([os.path.splitext(file_path)[0], "infos.txt"])
            return (
                # data url
                DataUrl(
                    file_path=url.file_path(),
                    scheme=scheme,
                ),
                # medata url
                DataUrl(
                    file_path=metadata_file,
                    scheme="ini",
                ),
            )

    @docstring(VolumeBase)
    def save_data(self, url: DataUrl | None = None) -> None:
        """
        :raises KeyError: if data path already exists and overwrite set to False
        :raises ValueError: if data is None
        """
        # to be discussed. Not sure we should raise an error in this case. Could be usefull but this could also be double edged knife
        if self.data is None:
            raise ValueError("No data to be saved")
        check_has_tiffle_file("raises")

        url = url or self.data_url
        if url is None:
            raise ValueError(
                "Cannot get data_url. An url should be provided. Don't know where to save this."
            )

        if url.scheme() == "tifffile":
            if url.data_path() is not None:
                raise ValueError("No data path expected. Unagleto save data")
            else:
                _logger.info(f"save data to {url.path()}")

            with TiffWriter(url.file_path(), bigtiff=True, append=self.append) as tif:
                if self.data.ndim == 2:
                    tif.write(self.data)
                elif self.data.ndim == 3:
                    for data_slice in self.data:
                        tif.write(data_slice)
                else:
                    raise ValueError(f"data should be 3D and not {self.data.ndim}D")

        else:
            raise ValueError(f"Scheme {url.scheme()} is not handled")

    @docstring(VolumeBase)
    def data_file_saver_generator(self, n_frames, data_url: DataUrl, overwrite: bool):
        """
        warning: the file will be open until the generator exists
        """

        class _FrameDumper:
            """
            will not work for VirtualLayout
            """

            def __init__(self, url, append) -> None:
                self.url = url
                self.append = append

            def __setitem__(self, key, value):
                if self.url.scheme() == "tifffile":
                    if self.url.data_path() is not None:
                        raise ValueError("No data path expected. Unagleto save data")
                    else:
                        _logger.info(f"save data to {self.url.path()}")

                    if key != slice(None, None, None):
                        raise ValueError("item setting only handle ':' for now")
                    with TiffWriter(
                        self.url.file_path(), bigtiff=True, append=self.append
                    ) as tif:
                        tif.write(value)
                else:
                    raise ValueError(f"Scheme {self.url.scheme()} is not handled")

        for i_frame in range(n_frames):
            yield _FrameDumper(data_url, append=self.append if i_frame == 0 else True)

    @docstring(VolumeBase)
    def save_metadata(self, url: DataUrl | None = None) -> None:
        """
        :raises KeyError: if data path already exists and overwrite set to False
        :raises ValueError: if data is None
        """
        if self.metadata is None:
            raise ValueError("No metadata to be saved")
        check_has_tiffle_file("raises")

        url = url or self.metadata_url
        if url is None:
            raise ValueError(
                "Cannot get metadata_url. An url should be provided. Don't know where to save this."
            )
        _logger.info(f"save metadata to {url.path()}")
        if url.scheme() == "ini":
            if url.data_path() is not None:
                raise ValueError("data_path is not handled by 'ini' scheme")
            else:
                dicttoini(
                    self.metadata,
                    url.file_path(),
                )
        else:
            raise ValueError(f"Scheme {url.scheme()} is not handled by multiframe tiff")

    @docstring(VolumeBase)
    def load_data(
        self, url: DataUrl | None = None, store: bool = True
    ) -> numpy.ndarray:
        url = url or self.data_url
        if url is None:
            raise ValueError(
                "Cannot get data_url. An url should be provided. Don't know where to save this."
            )

        data = numpy.asarray([slice for slice in self.browse_slices(url=url)])

        if store:
            self.data = data

        return data

    @docstring(VolumeBase)
    def load_metadata(self, url: DataUrl | None = None, store: bool = True) -> dict:
        url = url or self.metadata_url
        if url is None:
            raise ValueError(
                "Cannot get metadata_url. An url should be provided. Don't know where to save this."
            )

        if url.scheme() == "ini":
            metadata_file = url.file_path()
            if url.data_path() is not None:
                raise ValueError("data_path is not handled by ini scheme")
            else:
                try:
                    metadata = load_ini(metadata_file, "ini")
                except FileNotFoundError:
                    _logger.warning(f"unable to load metadata from {metadata_file}")
                    metadata = {}
        else:
            raise ValueError(f"Scheme {url.scheme()} is not handled by multiframe tiff")

        if store:
            self.metadata = metadata
        return metadata

    @staticmethod
    @docstring(VolumeBase)
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, MultiTiffVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {MultiTiffVolumeIdentifier}"
            )
        return MultiTIFFVolume(
            file_path=identifier.file_path,
        )

    @docstring(VolumeBase)
    def get_identifier(self) -> MultiTiffVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")
        return MultiTiffVolumeIdentifier(object=self, tiff_file=self.url.file_path())

    def browse_metadata_files(self, url=None):
        """
        return a generator go through all the existings files associated to the data volume
        """
        url = url or self.metadata_url
        if url is None:
            return
        elif url.file_path() is not None and os.path.exists(url.file_path()):
            yield url.file_path()

    def browse_data_files(self, url=None):
        """
        return a generator go through all the existings files associated to the data volume
        """
        url = url or self.data_url
        if url is None:
            return
        elif url.file_path() is not None and os.path.exists(url.file_path()):
            yield url.file_path()

    def browse_data_urls(self, url=None):
        url = url or self.data_url
        for data_file in self.browse_data_files(url=url):
            yield DataUrl(
                file_path=data_file,
                scheme=url.scheme(),
            )

    @docstring(VolumeBase)
    def browse_slices(self, url=None):
        if url is None and self.data is not None:
            for data_slice in self.data:
                yield data_slice
        else:
            url = url or self.data_url
            if url is None:
                raise ValueError(
                    "No data and data_url know and no url provided. Unable to browse slices"
                )
            if url.scheme() == "tifffile":
                if url.data_path() is not None:
                    raise ValueError("data_path is not handle by multiframe tiff")

                url = url or self.data_url
                reader = TiffFile(url.file_path())
                for series in reader.series:
                    data = series.asarray()
                    if data.ndim == 3:
                        for data_slice in data:
                            yield data_slice
                    elif data.ndim == 2:
                        yield data
                    else:
                        raise ValueError("series is expected to be 2D or 3D")
            else:
                raise ValueError(
                    f"Scheme {url.scheme()} is not handled by multiframe tiff"
                )

    def get_volume_shape(self, url=None):
        if self.data is not None:
            return self.data.shape
        url = url or self.data_url

        with tifffile.TiffFile(url.file_path()) as t:
            shapes = [series.shape for series in t.series]
            # assume that all series have the same dimensions for axis 1 and 2
            vol_shape = (len(t.series), shapes[0][0], shapes[0][1])
        return vol_shape

    def _get_tiff_volume_dtype(self):
        with tifffile.TiffFile(self.url.file_path()) as t:
            dtype = t.series[0].dtype
        # assume that dtype is the same for all series
        return dtype

    @docstring(VolumeBase)
    def load_chunk(self, chunk, url=None):
        vol_shape = self.get_volume_shape()
        vol_dtype = self._get_tiff_volume_dtype()
        chunk_shape = get_subvolume_shape(chunk, vol_shape)
        data_chunk = numpy.zeros(chunk_shape, dtype=vol_dtype)
        start_z = chunk[0].start or 0
        for i, image in enumerate(self.browse_slices(url=url)):
            if i >= start_z and i - start_z < chunk_shape[0]:
                data_chunk[i - start_z, ...] = image[chunk[1:]]
        return data_chunk

    @staticmethod
    def example_defined_from_str_identifier() -> str:
        return (
            MultiTIFFVolume(file_path="/path/to/tiff_file.tif")
            .get_identifier()
            .to_str()
        )



import os
import h5py
import numpy

from tomoscan.io import check_virtual_sources_exist


def test_vds(tmp_path):
    h5_file_without_vds = os.path.join(tmp_path, "h5_file_without_vds.hdf5")

    with h5py.File(h5_file_without_vds, mode="w") as h5f:
        h5f["data"] = numpy.random.random((120, 120))
    assert check_virtual_sources_exist(h5_file_without_vds, "data")

    h5_file_with_vds = os.path.join(tmp_path, "h5_file_with_vds.hdf5")

    # create some dataset
    for i in range(4):
        filename = os.path.join(tmp_path, f"{i}.h5")
        with h5py.File(filename, mode="w") as h5f:
            h5f.create_dataset("data", (100,), "i4", numpy.arange(100))

    layout = h5py.VirtualLayout(shape=(4, 100), dtype="i4")
    for i in range(4):
        filename = os.path.join(tmp_path, f"{i}.h5")
        layout[i] = h5py.VirtualSource(filename, "data", shape=(100,))

    with h5py.File(h5_file_with_vds, mode="w") as h5f:
        # create the virtual dataset
        h5f.create_virtual_dataset("data", layout, fillvalue=-5)
    assert check_virtual_sources_exist(h5_file_with_vds, "data")



# coding: utf-8
"""module defining utils for an .vol volume (also know as raw)"""

from __future__ import annotations

import logging
import os
import sys
from xml.dom.minidom import parseString as parse_xml_string

import h5py
import numpy
from dicttoxml import dicttoxml
from silx.io.dictdump import dicttoini
from silx.io.url import DataUrl

from tomoscan.esrf.identifier.rawidentifier import RawVolumeIdentifier
from tomoscan.scanbase import TomoScanBase
from tomoscan.utils import docstring
from tomoscan.volumebase import VolumeBase

_logger = logging.getLogger(__name__)

__all__ = ["RawVolume"]


class RawVolume(VolumeBase):
    """
    Volume where data si saved under .vol binary file and metadata are saved in .vol.info and or .vol.xml
    Note: for now reading information from the .xml is not managed. We expect to write one or both and read from the text file (.vol.info)
    Warning: meant as legacy for pyhst .vol file and existing post processing tool. We mostly expect software to write .vol file.
    """

    DEFAULT_DATA_SCHEME = "raw"

    DEFAULT_DATA_EXTENSION = "vol"

    DEFAULT_METADATA_SCHEME = "info"

    DEFAULT_METADATA_EXTENSION = "vol.info"

    def __init__(
        self,
        file_path: str | None = None,
        data: numpy.ndarray | None = None,
        source_scan: TomoScanBase | None = None,
        metadata: dict | None = None,
        data_url: DataUrl | None = None,
        metadata_url: DataUrl | None = None,
        overwrite: bool = False,
        append: bool = False,
        data_extension=DEFAULT_DATA_EXTENSION,
        metadata_extension=DEFAULT_METADATA_EXTENSION,
    ) -> None:
        if file_path is not None:
            url = DataUrl(file_path=file_path, data_path=None, scheme="raw")
        else:
            url = None

        self._file_path = file_path
        super().__init__(
            url=url,
            data=data,
            source_scan=source_scan,
            metadata=metadata,
            data_url=data_url,
            metadata_url=metadata_url,
            overwrite=overwrite,
            data_extension=data_extension,
            metadata_extension=metadata_extension,
        )
        self.append = append

    @property
    def data_extension(self):
        if self.data_url is not None and self.data_url.file_path() is not None:
            return os.path.splitext(self.data_url.file_path())[1]

    @property
    def metadata_extension(self):
        if self.metadata_url is not None and self.metadata_url.file_path() is not None:
            return os.path.splitext(self.metadata_url.file_path())[1]

    @VolumeBase.data.setter
    def data(self, data):
        if not isinstance(data, (numpy.ndarray, type(None), h5py.VirtualLayout)):
            raise TypeError(
                f"data is expected to be None or a numpy array not {type(data)}"
            )
        if isinstance(data, numpy.ndarray) and data.ndim != 3:
            raise ValueError(f"data is expected to be 3D and not {data.ndim}D.")
        self._data = data

    @property
    def file_path(self):
        return self._file_path

    @file_path.setter
    def file_path(self, file_path: str | None):
        if not (file_path is None or isinstance(file_path, str)):
            raise TypeError
        self._file_path = file_path
        self.url = DataUrl(file_path=file_path, data_path=None, scheme="raw")

    @docstring(VolumeBase)
    def deduce_data_and_metadata_urls(self, url: DataUrl | None) -> tuple:
        if url is None:
            return None, None
        else:
            if url.data_slice() is not None:
                raise ValueError(f"data_slice is not handled by the {RawVolume}")
            file_path = url.file_path()
            data_path = url.data_path()
            if data_path is not None:
                raise ValueError("data_path is not handle by the .vol volume.")
            scheme = url.scheme() or "raw"
            metadata_info_file = os.path.splitext(url.file_path())[0] + ".vol.info"
            return (
                # data url
                DataUrl(
                    file_path=file_path,
                    data_path=None,
                    scheme=scheme,
                ),
                # medata url
                DataUrl(
                    file_path=metadata_info_file,
                    data_path=None,
                    scheme=self.DEFAULT_METADATA_SCHEME,
                ),
            )

    @docstring(VolumeBase)
    def save_data(self, url: DataUrl | None = None, **kwargs) -> None:
        if self.data is None:
            return
        url = url or self.data_url
        if url is None:
            raise ValueError(
                "Cannot get data_url. An url should be provided. Don't know where to save this."
            )

        if url.scheme() != "raw":
            raise ValueError("Unsupported scheme - please use scheme='raw'")
        if url.data_path() is not None:
            raise ValueError("No data path expected. Unagleto save data")

        _logger.info(f"save data to {url.path()}")

        if self.data.dtype != numpy.float32:
            raise TypeError(".vol format only takes float32 as data type")

        # check endianness: make sure data is lowbytefirst
        if self.data.dtype.byteorder == ">" or (
            self.data.dtype.byteorder == "=" and sys.byteorder != "little"
        ):
            # lowbytefirst
            raise TypeError("data is expected to be byteorder: low byte first")

        if self.data.ndim == 3:
            data = self.data
        elif self.data.ndim == 2:
            data = self.data.reshape(1, self.data.shape[0], self.data.shape[1])
        else:
            raise ValueError(f"data should be 3D and not {self.data.ndim}D")

        file_mode = "ab" if self.append else "wb"
        with open(url.file_path(), file_mode) as fdesc:
            if self.append:
                n_bytes = os.path.getsize(url.file_path())
                fdesc.seek(n_bytes)
            data.tofile(fdesc)

    @docstring(VolumeBase)
    def load_data(self, url: DataUrl | None = None, store: bool = True) -> numpy.array:
        url = url or self.data_url
        if url is None:
            raise ValueError(
                "Cannot get data_url. An url should be provided. Don't know where to save this."
            )

        if self.metadata is None:
            # for .vol file we need metadata to get shape - expected in a .vol.info file
            metadata = self.load_metadata(store=False)
        else:
            metadata = self.metadata

        dimX = metadata.get("NUM_X", None)
        dimY = metadata.get("NUM_Y", None)
        dimZ = metadata.get("NUM_Z", None)
        byte_order = metadata.get("BYTEORDER", "LOWBYTEFIRST")
        if byte_order.lower() == "highbytefirst":
            byte_order = ">"
        elif byte_order.lower() == "lowbytefirst":
            byte_order = "<"
        else:
            raise ValueError(f"Unable to interpret byte order value: {byte_order}")

        if dimX is None or dimY is None or dimZ is None:
            _logger.error(f"Unable to get volume shape (get: {dimZ, dimY, dimZ} )")
            data = None
        else:
            shape = (int(dimZ), int(dimY), int(dimX))
            try:
                data_type = numpy.dtype(byte_order + "f")
                data = numpy.fromfile(
                    url.file_path(), dtype=data_type, count=-1, sep=""
                )
            except Exception as e:
                _logger.warning(
                    f"Fail to load data from {url.file_path()}. Error is {e}."
                )
                data = None
            else:
                data = data.reshape(shape)
                if store is True:
                    self.data = data
        return data

    @docstring(VolumeBase)
    def save_metadata(self, url: DataUrl | None = None, store: bool = True) -> None:
        """
        :raises KeyError: if data path already exists and overwrite set to False
        :raises ValueError: if data is None
        """
        if self.metadata is None:
            raise ValueError("No metadata to be saved")
        url = url or self.metadata_url
        if url is None:
            raise ValueError(
                "Cannot get metadata_url. An url should be provided. Don't know where to save this."
            )
        _logger.info(f"save metadata to {url.path()}")
        if url.scheme() == "info":
            metadata_file = url.file_path()
            _logger.info(f"save data to {metadata_file}")
            if len(self.metadata) > 0:
                # same as ini but no section. Write works but read fails
                dicttoini(self.metadata, metadata_file)
        elif url.scheme() == "lxml":
            metadata_file = url.file_path()
            _logger.info(f"save data to {metadata_file}")
            if len(self.metadata) > 0:
                # Format metadata to a XML file, with a format that can be read by imagej.
                # Does not make sense to you ? For us neither!
                size_xyz = [
                    int(self.metadata.get(key, 0))
                    for key in ["NUM_X", "NUM_Y", "NUM_Z"]
                ]
                if size_xyz == 0:
                    _logger.error(
                        "Something wrong with NUM_X, NUM_Y or NUM_X: missing or zero ?"
                    )
                metadata_for_xml = {
                    "reconstruction": {
                        "idAc": "N_A_",
                        "listSubVolume": {
                            "subVolume": {
                                "SUBVOLUME_NAME": os.path.basename(
                                    self.data_url.file_path()
                                ),
                                "SIZEX": size_xyz[0],
                                "SIZEY": size_xyz[1],
                                "SIZEZ": size_xyz[2],
                                "ORIGINX": 1,
                                "ORIGINY": 1,
                                "ORIGINZ": 1,
                                "DIM_REC": numpy.prod(size_xyz),
                                "BYTE_ORDER": "LOWBYTEFIRST",  # !
                            }
                        },
                    }
                }
                for what in ["voxelSize", "ValMin", "ValMax", "s1", "s2", "S1", "S2"]:
                    metadata_for_xml["reconstruction"]["listSubVolume"]["subVolume"][
                        what
                    ] = float(self.metadata.get(what, 0.0))

                xml_str = dicttoxml(
                    metadata_for_xml,
                    custom_root="tomodb2",
                    xml_declaration=False,
                    attr_type=False,
                    return_bytes=False,
                )
                xml_str_pretty = parse_xml_string(xml_str).toprettyxml(indent="  ")

                with open(metadata_file, mode="w") as file_:
                    file_.write(xml_str_pretty)
        else:
            raise ValueError(f"scheme {url.scheme()} is not handled")

    @docstring(VolumeBase)
    def load_metadata(self, url: DataUrl | None = None, store: bool = True) -> dict:
        url = url or self.metadata_url
        if url is None:
            raise ValueError(
                "Cannot get metadata_url. An url should be provided. Don't know where to save this."
            )
        if url.scheme() == "info":

            def info_file_to_dict(info_file):
                ddict = {}
                with open(info_file, "r") as _file:
                    lines = _file.readlines()
                    for line in lines:
                        if "=" not in line:
                            continue
                        _line = line.rstrip().replace(" ", "")
                        _line = _line.split("#")[0]
                        key, value = _line.split("=")
                        ddict[key] = value
                return ddict

            metadata_file = url.file_path()
            if url.data_path() is not None:
                raise ValueError("data_path is not handled by ini scheme")
            else:
                try:
                    metadata = info_file_to_dict(metadata_file)
                except FileNotFoundError:
                    _logger.warning(f"unable to load metadata from {metadata_file}")
                    metadata = {}
        else:
            raise ValueError(f"scheme {url.scheme()} is not handled")
        if store:
            self.metadata = metadata
        return metadata

    def get_volume_shape(self, url=None):
        metadata = self.metadata or self.load_metadata()
        from_metadata = (
            metadata.get("SIZEZ"),
            metadata.get("SIZEY"),
            metadata.get("SIZEX"),
        )
        if (
            from_metadata[0] is not None
            and from_metadata[1] is not None
            and from_metadata[2] is not None
        ):
            return from_metadata
        else:
            if self.data is None:
                data = self.load_data(url=url)
            else:
                data = self.data
            if data is not None:
                return data.shape
            else:
                return None

    def browse_metadata_files(self, url=None):
        """
        return a generator go through all the existings files associated to the data volume
        """
        url = url or self.metadata_url
        if url is None:
            return
        elif url.file_path() is not None and os.path.exists(url.file_path()):
            yield url.file_path()

    def browse_data_files(self, url=None):
        """
        return a generator go through all the existings files associated to the data volume
        """
        url = url or self.data_url
        if url is None:
            return
        elif url.file_path() is not None and os.path.exists(url.file_path()):
            yield url.file_path()

    def browse_slices(self, url=None):
        if url is not None or self.data is None:
            data = self.load_data(url=url)
        else:
            data = self.data

        for vol_slice in data:
            yield vol_slice

    def browse_data_urls(self, url=None):
        url = url or self.data_url
        if url is not None and os.path.exists(url.file_path()):
            yield url

    @docstring(VolumeBase)
    def data_file_saver_generator(
        self, n_frames, data_url: DataUrl, overwrite: bool, mode: str = "a", **kwargs
    ):
        """
        warning: the file will be open until the generator exists
        """

        class _FrameDumper:
            """
            will not work for VirtualLayout
            """

            Dataset = None
            # shared dataset

            def __init__(
                self,
                fid,
            ) -> None:
                self._fid = fid

            def __setitem__(self, key, value):
                if key != slice(None, None, None):
                    raise ValueError("item setting only handle ':' for now")
                if not isinstance(value, numpy.ndarray):
                    raise TypeError(
                        "value is expected to be an instance of numpy.ndarray"
                    )
                value.tofile(self._fid)

        if (
            data_url.file_path() is not None
            and os.path.dirname(data_url.file_path()) != ""
        ):
            os.makedirs(os.path.dirname(data_url.file_path()), exist_ok=True)

        with open(self.data_url.file_path(), "wb") as fid:
            for _ in range(n_frames):
                yield _FrameDumper(fid=fid)

    @staticmethod
    @docstring(VolumeBase)
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, RawVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {RawVolumeIdentifier}"
            )
        return RawVolume(
            file_path=identifier.file_path,
        )

    @docstring(VolumeBase)
    def get_identifier(self) -> RawVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")
        return RawVolumeIdentifier(object=self, file_path=self.url.file_path())

    @staticmethod
    def example_defined_from_str_identifier() -> str:
        """example as string to explain how users can defined identifiers from a string"""
        return " ; ".join(
            [
                f"{RawVolume(file_path='/path/to/my/my_volume.vol').get_identifier().to_str()}",
            ]
        )



from posixpath import join
from .h5utils import get_first_hdf5_entry, get_h5_value as _get_h5_value


def get_h5_value(fname, h5_path, default_ret=None):
    val = _get_h5_value(fname, h5_path, default_ret=default_ret)
    if isinstance(val, bytes):
        val = val.decode()
    return val


def get_title(fname, entry):
    title = get_h5_value(fname, join(entry, "title"))
    if isinstance(title, bytes):
        title = title.decode()
    return title


class ScanTypeBase:
    """
    Type scan base
    """

    default_detector_name = "pilatus"
    required_fields = []
    optional_fields = []

    def __init__(self, fname, entry=None, detector_name=None):
        self.fname = fname
        self.detector_name = detector_name or self.default_detector_name
        self.entry = entry or get_first_hdf5_entry(fname)
        self.title = get_title(self.fname, self.entry)

    def get_motors(self):
        raise NotImplementedError("Base class")

    def _handle_field_not_found(self, field, on_error):
        msg = "Cannot find field %s in entry %s of file %s" % (
            field,
            self.entry,
            self.fname,
        )
        if on_error == "raise":
            raise ValueError(msg)
        elif on_error == "print":
            print(msg)
        # else: pass

    def _format_variable_field(self, field):
        if "{" not in field:
            return field
        if "{motor}" in field:
            motors = self.get_motors()
            field_formatted = field.format(motor=motors)
        elif "{detector}" in field:
            field_formatted = field.format(detector=self.detector_name)
        else:
            raise ValueError("Unsupported variable: %s" % field)
        return field_formatted

    def get_metadata(self, on_error="raise"):
        """
        Check that all the metadata necessary to carry on XRD tomography reconstruction is present.
        Return a dictionary with the associated metadata.
        """

        def _add_field(field, metadata, on_error):
            field = self._format_variable_field(field)
            field_path = join(self.entry, field)
            val = get_h5_value(self.fname, field_path)
            if val is None:
                self._handle_field_not_found(field, on_error)
            metadata[field] = val

        metadata = {}
        for field in self.required_fields:
            _add_field(field, metadata, on_error)
        for field in self.optional_fields:
            _add_field(field, metadata, "ignore")
        return metadata


class LimaTake(ScanTypeBase):
    required_fields = [
        "start_time",
        "end_time",
        "title",
        "instrument/{detector}/acq_parameters/acq_expo_time",
    ]
    optional_fields = []


class LoopScan(ScanTypeBase):
    required_fields = LimaTake.required_fields + [
        "measurement/fpico2",
        "measurement/fpico3",
        "measurement/epoch",
        "measurement/elapsed_time",
    ]
    optional_fields = [
        "measurement/{detector}_roi1" + x for x in ["", "_avg", "_max", "_min", "_std"]
    ]


FtimeScan = LoopScan


class FScan(ScanTypeBase):
    required_fields = LimaTake.required_fields + [
        "measurement/fpico2",
        "measurement/fpico3",
        "measurement/epoch_trig",
        "measurement/{motor}",  # motor name in title, more than one in case of fscan2d
    ]
    optional_fields = LoopScan.optional_fields

    def get_motors(self):
        fscan_info = self.title.split()
        if fscan_info[0].lower() != "fscan":
            raise ValueError(
                "Not a fscan: %s in file %s entry %s"
                % (self.title, self.fname, self.entry)
            )
        return fscan_info[1]  # list to be consistent with FScan2D ?


class FScan2D(ScanTypeBase):
    required_fields = FScan.required_fields
    optional_fields = FScan.optional_fields

    def get_motors(self):
        fscan_info = self.title.split()
        if fscan_info[0].lower() != "fscan2d":
            raise ValueError(
                "Not a fscan2d: %s in file %s entry %s"
                % (self.title, self.fname, self.entry)
            )
        # This is still too brittle.
        # ID11 FSCAN2D entries are in the form "fscan2d dty -300 1 1 rot -90 0.05 3620 0.002 0.00200017"
        ret = fscan_info[1:2]
        if isinstance(ret, list):
            ret = ret[0]
        return ret


class AeroystepScan(ScanTypeBase):
    required_fields = LimaTake.required_fields + [
        "measurement/fpico2",
        "measurement/fpico3",
        "measurement/epoch_trig",
        "measurement/hrrz",
        "measurement/hry",
        "instrument/positioners/hrz",
    ]
    optional_fields = LoopScan.optional_fields


scan_classes = {
    "limatake": LimaTake,
    "loopscan": LoopScan,
    "ftimescan": FtimeScan,
    "fscan": FScan,
    "fscan2d": FScan2D,
    "aeroystepscan": AeroystepScan,
}


# class factory
def Scan(fname, entry=None, detector_name=None, raise_error_if_not_supported=True):
    entry = entry or get_first_hdf5_entry(fname)
    title = get_title(fname, entry)
    title = title.split(" ")[0]  # handle stuff like "loopscan 1 1"
    scan_cls = scan_classes.get(title, None)
    if scan_cls is None:
        if raise_error_if_not_supported:
            raise ValueError(
                "Unsupported scan type '%s' - supported are %s"
                % (title, str(list(scan_classes.keys())))
            )
        return None
    return scan_cls(fname, entry=entry, detector_name=detector_name)


def get_scan_metadata(fname, entry=None, detector_name=None, on_error="raise"):
    scan = Scan(
        fname,
        entry=entry,
        detector_name=detector_name,
        raise_error_if_not_supported=False,
    )
    if scan is None:
        return None
    return scan.get_metadata(on_error=on_error)



from __future__ import annotations


import numpy
from typing import Iterable


class ReducedFramesInfos:
    """contains reduced frames metadata as count_time and machine_electric_current"""

    MACHINE_ELECT_CURRENT_KEY = "machine_electric_current"

    COUNT_TIME_KEY = "count_time"

    def __init__(self) -> None:
        self._count_time = []
        self._machine_electric_current = []

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, dict):
            return ReducedFramesInfos().load_from_dict(__o) == self
        if not isinstance(__o, ReducedFramesInfos):
            return False
        return numpy.array_equal(
            numpy.array(self.count_time), numpy.array(__o.count_time)
        ) and numpy.array_equal(
            numpy.array(self.machine_electric_current),
            numpy.array(__o.machine_electric_current),
        )

    def clear(self):
        self._count_time.clear()
        self._machine_electric_current.clear()

    @property
    def count_time(self) -> list:
        """
        frame exposure time in second
        """
        return self._count_time

    @count_time.setter
    def count_time(self, count_time: Iterable | None):
        if count_time is None:
            self._count_time.clear()
        else:
            self._count_time = list(count_time)

    @property
    def machine_electric_current(self) -> list:
        """
        machine electric current in Ampere
        """
        return self._machine_electric_current

    @machine_electric_current.setter
    def machine_electric_current(self, machine_electric_current: Iterable | None):
        if machine_electric_current is None:
            self._machine_electric_current.clear()
        else:
            self._machine_electric_current = list(machine_electric_current)

    def to_dict(self) -> dict:
        res = {}
        if len(self.machine_electric_current) > 0:
            res[self.MACHINE_ELECT_CURRENT_KEY] = self.machine_electric_current
        if len(self.count_time) > 0:
            res[self.COUNT_TIME_KEY] = self.count_time
        return res

    def load_from_dict(self, my_dict: dict):
        self.machine_electric_current = my_dict.get(
            self.MACHINE_ELECT_CURRENT_KEY, None
        )
        self.count_time = my_dict.get(self.COUNT_TIME_KEY, None)
        return self

    @staticmethod
    def pop_info_keys(my_dict: dict):
        if not isinstance(my_dict, dict):
            raise TypeError
        my_dict.pop(ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY, None)
        my_dict.pop(ReducedFramesInfos.COUNT_TIME_KEY, None)
        return my_dict

    @staticmethod
    def split_data_and_metadata(my_dict):
        metadata = ReducedFramesInfos().load_from_dict(my_dict)
        data = ReducedFramesInfos.pop_info_keys(my_dict)
        return data, metadata

    def __str__(self):
        return "\n".join(
            [
                f"machine_electric_current {self.machine_electric_current}",
                f"count_time {self.count_time}",
            ]
        )



# coding: utf-8
"""Module containing the TomoObject class. Parent class of any tomo object"""

from __future__ import annotations

from tomoscan.utils import BoundingBox1D

from .identifier import BaseIdentifier

__all__ = [
    "TomoObject",
]


class TomoObject:
    """Parent class of all tomographic object in tomoscan"""

    @staticmethod
    def from_identifier(identifier: str | BaseIdentifier):
        """Return the Dataset from a identifier"""
        raise NotImplementedError("Base class")

    def get_identifier(self) -> BaseIdentifier:
        """dataset unique identifier. Can be for example a hdf5 and
        en entry from which the dataset can be rebuild"""
        raise NotImplementedError("Base class")

    def get_bounding_box(self, axis: str | int | None = None) -> BoundingBox1D:
        """
        Return the bounding box covered by the Tomo object
        axis is expected to be in (0, 1, 2) or (x==0, y==1, z==2)
        """
        raise NotImplementedError("Base class")

    def build_drac_metadata(self) -> dict:
        """
        build icat metadata dictionary filling NXtomo definition following icat definition: https://gitlab.esrf.fr/icat/hdf5-master-config/-/blob/88a975039694d5dba60e240b7bf46c22d34065a0/hdf5_cfg.xml
        """
        raise NotImplementedError()



from __future__ import annotations

from silx.utils.enum import Enum as _Enum


class SourceType(_Enum):
    SPALLATION_NEUTRON = "Spallation Neutron Source"
    PULSED_REACTOR_NEUTRON_SOURCE = "Pulsed Reactor Neutron Source"
    REACTOR_NEUTRON_SOURCE = "Reactor Neutron Source"
    SYNCHROTRON_X_RAY_SOURCE = "Synchrotron X-ray Source"
    PULSED_MUON_SOURCE = "Pulsed Muon Source"
    ROTATING_ANODE_X_RAY = "Rotating Anode X-ray"
    FIXED_TUBE_X_RAY = "Fixed Tube X-ray"
    UV_LASER = "UV Laser"
    FREE_ELECTRON_LASER = "Free-Electron Laser"
    OPTICAL_LASER = "Optical Laser"
    ION_SOURCE = "Ion Source"
    UV_PLASMA_SOURCE = "UV Plasma Source"
    METAL_JET_X_RAY = "Metal Jet X-ray"


class Source:
    """Information regarding the x-ray storage ring/facility"""

    def __init__(self, name=None, type=None):
        self._name = name
        self._type = type

    @property
    def name(self) -> str | None:
        return self._name

    @name.setter
    def name(self, name: str | None):
        if not isinstance(name, (str, type(None))):
            raise TypeError("name is expected to be None or a str")
        self._name = name

    @property
    def type(self) -> SourceType | None:
        return self._type

    @type.setter
    def type(self, type_: str | SourceType | None):
        if type_ is None:
            self._type = None
        else:
            type_ = SourceType.from_value(type_)
            self._type = type_

    def __str__(self):
        return f"source (name: {self.name}, type: {self.type})"



# coding: utf-8
"""Module containing validators"""

import pytest

from tomoscan.volumebase import VolumeBase


def test_volume_base():
    """Test VolumeBase file"""

    class UnplemetnedVolumeBase(VolumeBase):
        def deduce_data_and_metadata_urls(self, url):
            return None, None

    volume_base = UnplemetnedVolumeBase()

    with pytest.raises(NotImplementedError):
        volume_base.example_defined_from_str_identifier()

    with pytest.raises(NotImplementedError):
        volume_base.get_identifier()

    with pytest.raises(NotImplementedError):
        VolumeBase.from_identifier("")

    with pytest.raises(NotImplementedError):
        volume_base.save_data()

    with pytest.raises(NotImplementedError):
        volume_base.save_metadata()

    with pytest.raises(NotImplementedError):
        volume_base.save()

    with pytest.raises(NotImplementedError):
        volume_base.load_data()

    with pytest.raises(NotImplementedError):
        volume_base.load_metadata()

    with pytest.raises(NotImplementedError):
        volume_base.load()

    with pytest.raises(NotImplementedError):
        volume_base.browse_data_files()

    with pytest.raises(NotImplementedError):
        volume_base.browse_metadata_files()

    with pytest.raises(NotImplementedError):
        volume_base.browse_data_urls()

    volume_base.position = (0, 1, 2)
    assert isinstance(volume_base.position, tuple)
    assert volume_base.position == (0, 1, 2)

    volume_base.voxel_size = (12.3, 2.5, 6.9)
    assert volume_base.voxel_size == (12.3, 2.5, 6.9)



# coding: utf-8

import pytest

from tomoscan.utils.geometry import BoundingBox1D, BoundingBox3D, _BoundingBox


def test_bounding_box_base():
    bb = _BoundingBox(0, 1)
    with pytest.raises(NotImplementedError):
        bb.get_overlap(None)


def test_bounding_box_1D():
    """
    check if BoundingBox1D is working properly
    """
    # check overlaping
    bb1 = BoundingBox1D(0.0, 1.0)
    bb2 = BoundingBox1D(0.2, 1.0)
    assert bb1.get_overlap(bb2) == BoundingBox1D(0.2, 1.0)
    assert bb2.get_overlap(bb1) == BoundingBox1D(0.2, 1.0)

    bb1 = BoundingBox1D(0.0, 1.0)
    bb2 = BoundingBox1D(0.2, 0.8)
    assert bb1.get_overlap(bb2) == BoundingBox1D(0.2, 0.8)
    assert bb2.get_overlap(bb1) == BoundingBox1D(0.2, 0.8)

    bb1 = BoundingBox1D(0.0, 1.0)
    bb2 = BoundingBox1D(1.0, 1.2)
    assert bb2.get_overlap(bb1) == BoundingBox1D(1.0, 1.0)

    # check outside
    bb1 = BoundingBox1D(0.0, 1.0)
    bb2 = BoundingBox1D(2.0, 2.2)

    assert bb2.get_overlap(bb1) is None
    assert bb1.get_overlap(bb2) is None

    # check on fully including in the other
    bb1 = BoundingBox1D(0.0, 1.0)
    bb2 = BoundingBox1D(0.1, 0.3)
    assert bb2.get_overlap(bb1) == BoundingBox1D(0.1, 0.3)
    assert bb1.get_overlap(bb2) == BoundingBox1D(0.1, 0.3)

    with pytest.raises(TypeError):
        bb1.get_overlap(None)


def test_bounding_box_3D():
    """
    check if BoundingBox3D is working properly
    """
    # check overlaping
    bb1 = BoundingBox3D((0.0, -0.1, 0.0), [1.0, 0.8, 0.9])
    bb2 = BoundingBox3D([0.2, 0.0, 0.1], (1.0, 2.0, 3.0))
    assert bb1.get_overlap(bb2) == BoundingBox3D((0.2, 0.0, 0.1), (1.0, 0.8, 0.9))
    assert bb2.get_overlap(bb1) == BoundingBox3D((0.2, 0.0, 0.1), (1.0, 0.8, 0.9))

    # check outside
    bb1 = BoundingBox3D((0.0, -0.1, 0.0), [1.0, 0.8, 0.9])
    bb2 = BoundingBox3D([0.2, 0.0, -2.1], (1.0, 2.0, -1.0))

    assert bb2.get_overlap(bb1) is None
    assert bb1.get_overlap(bb2) is None

    # check on fully including in the other
    bb1 = BoundingBox3D((0.0, 0.1, 0.2), (1.0, 1.1, 1.2))
    bb2 = BoundingBox3D((-2.0, -3.0, -4.0), (2.0, 2.0, 2.0))
    assert bb2.get_overlap(bb1) == BoundingBox3D((0.0, 0.1, 0.2), (1.0, 1.1, 1.2))
    assert bb1.get_overlap(bb2) == BoundingBox3D((0.0, 0.1, 0.2), (1.0, 1.1, 1.2))

    with pytest.raises(TypeError):
        bb1.get_overlap(None)



# coding: utf-8
"""utils function for esrf volumes"""

from __future__ import annotations

import logging
import os

import h5py
from silx.io.utils import open as hdf5_open

from tomoscan.volumebase import SliceTuple

from tomoscan.esrf.identifier.edfidentifier import EDFVolumeIdentifier
from tomoscan.esrf.identifier.hdf5Identifier import HDF5VolumeIdentifier
from tomoscan.esrf.identifier.jp2kidentifier import JP2KVolumeIdentifier
from tomoscan.esrf.identifier.rawidentifier import RawVolumeIdentifier
from tomoscan.esrf.identifier.tiffidentifier import (
    MultiTiffVolumeIdentifier,
    TIFFVolumeIdentifier,
)
from tomoscan.esrf.volume.edfvolume import EDFVolume
from tomoscan.esrf.volume.hdf5volume import HDF5Volume
from tomoscan.esrf.volume.jp2kvolume import JP2KVolume
from tomoscan.esrf.volume.rawvolume import RawVolume
from tomoscan.esrf.volume.tiffvolume import MultiTIFFVolume, TIFFVolume

_logger = logging.getLogger(__name__)

__all__ = [
    "guess_hdf5_volume_data_paths",
    "guess_volumes",
    "get_most_common_extension",
    "group_slices_by_axis",
]

_DEFAULT_SCHEME_TO_VOL = {
    EDFVolumeIdentifier.scheme: EDFVolume,
    HDF5VolumeIdentifier.scheme: HDF5Volume,
    TIFFVolumeIdentifier.scheme: TIFFVolume,
    MultiTiffVolumeIdentifier.scheme: MultiTIFFVolume,
    JP2KVolumeIdentifier.scheme: JP2KVolume,
    RawVolumeIdentifier.scheme: RawVolume,
}


def guess_hdf5_volume_data_paths(file_path, data_path="/", depth=3) -> tuple:
    """
    browse hdf5 file 'file_path' from 'data_path' on 'depth' level and check for possible defined volumes.

    :param file_path: file path to the hdf5 file to browse
    :param data_path: path in the file to start research
    :param depth: on which layer we should apply research
    :return: tuple of data_path that could fit a volume
    """
    if not h5py.is_hdf5(file_path):
        raise ValueError(f"{file_path} is not a hdf5 file path")
    with hdf5_open(filename=file_path) as h5f:
        group = h5f[data_path]
        if isinstance(group, h5py.Group):
            if HDF5Volume.DATA_DATASET_NAME in group:
                return (data_path,)
            elif depth > 0:
                res = []
                for key in group.keys():
                    res.extend(
                        guess_hdf5_volume_data_paths(
                            file_path=file_path,
                            data_path="/".join((data_path, key)).replace("//", "/"),
                            depth=depth - 1,
                        )
                    )
                return tuple(res)

        return tuple()


def guess_volumes(
    path, scheme_to_vol: dict | None = None, filter_histograms=True
) -> tuple:
    """
    from a file path or a folder path try to guess volume(s)

    :param path: file or folder path
    :param scheme_to_vol: dict to know which constructor to call. Key if the scheme, value if the volume constructor.
                               usefull for libraries redefining volume or adding some like tomwer.
                               If none provided will take the tomoscan default one
    :return: tuple of volume
    """
    if not os.path.exists(path):
        raise OSError("path doesn't exists")

    if scheme_to_vol is None:
        scheme_to_vol = _DEFAULT_SCHEME_TO_VOL

    if os.path.isfile(path):
        if h5py.is_hdf5(path):
            res = []
            for data_path in guess_hdf5_volume_data_paths(path):
                assert isinstance(data_path, str)
                res.append(
                    scheme_to_vol[HDF5VolumeIdentifier.scheme](
                        file_path=path,
                        data_path=data_path,
                    )
                )
            # filter potential 'nabu histogram'
            # as nabu histograms looks like volume simply look at the name
            # could also be on data ndim
            if filter_histograms is True:

                def is_not_histogram(vol_identifier):
                    return not (
                        hasattr(vol_identifier, "data_path")
                        and vol_identifier.data_path.endswith("histogram")
                    )

                res = tuple(filter(is_not_histogram, res))

            return tuple(res)
        elif path.lower().endswith((".tif", ".tiff")):
            return (scheme_to_vol[MultiTiffVolumeIdentifier.scheme](file_path=path),)
        elif path.lower().endswith((".vol", ".raw")):
            return (scheme_to_vol[RawVolumeIdentifier.scheme](file_path=path),)
    elif os.path.isdir(path):
        most_common_extension = get_most_common_extension(path)
        if most_common_extension is None:
            return tuple()

        basename = _guess_volume_basename(path, extension=most_common_extension)
        if most_common_extension in ("tiff", "tif"):
            return (
                scheme_to_vol[TIFFVolumeIdentifier.scheme](
                    folder=path,
                    volume_basename=basename,
                    data_extension=most_common_extension,
                ),
            )
        elif most_common_extension in ("jp2", "jp2k"):
            return (
                scheme_to_vol[JP2KVolumeIdentifier.scheme](
                    folder=path,
                    volume_basename=basename,
                    data_extension=most_common_extension,
                ),
            )
        elif most_common_extension == "edf":
            return (
                scheme_to_vol[EDFVolumeIdentifier.scheme](
                    folder=path,
                    volume_basename=basename,
                    data_extension=most_common_extension,
                ),
            )
        else:
            _logger.warning(
                f"most common extension is {most_common_extension}. Unable to create a volume from it"
            )
            return tuple()
    else:
        raise NotImplementedError("guess_volumes only handle file and folder...")


def get_most_common_extension(folder_path):
    if not os.path.isdir(folder_path):
        raise ValueError(f"a folder path is expected. {folder_path} isn't")

    extensions = {}
    for file_path in os.listdir(folder_path):
        _, ext = os.path.splitext(file_path)
        ext = ext.lower().lstrip(".")
        if ext in extensions:
            extensions[ext] += 1
        else:
            extensions[ext] = 1

    # filter not handled extensions
    def is_valid_extension(extension):
        return extension in ("edf", "tif", "tiff", "jp2", "jp2k")

    extensions = {
        key: value for (key, value) in extensions.items() if is_valid_extension(key)
    }

    if len(extensions) == 0:
        _logger.warning(f"no valid extensions found in {folder_path}")
    else:
        sort_extensions = sorted(extensions.items(), key=lambda x: x[1], reverse=True)
        return sort_extensions[0][0]


def _guess_volume_basename(folder_path, extension):
    # list all the files matching the file and guessing the file parttern
    files_to_check = []
    possible_basenames = {}
    for file_path in os.listdir(folder_path):
        if file_path.lower().endswith(extension):
            files_to_check.append(os.path.splitext(file_path)[0])
            # the expected way to save those files is basename_XXXX with XXXX is the index over 4 char
            basename = "_".join(file_path.split("_")[:-1])
            if basename in possible_basenames:
                possible_basenames[basename] += 1
            else:
                possible_basenames[basename] = 1

    if len(possible_basenames) == 0:
        _logger.warning(f"no valid basename found in {folder_path}")
    else:
        sort_basenames = sorted(
            possible_basenames.items(), key=lambda x: x[1], reverse=True
        )
        if len(sort_basenames) > 1:
            _logger.warning(
                f"more than one basename found. Take the most probable one ({sort_basenames[0][0]})"
            )
        return sort_basenames[0][0]


def group_slices_by_axis(
    slices: tuple[SliceTuple], volume_shape: tuple[int] | None
) -> dict[int, set]:
    """
    group a tuple of slices provided as SliceTuple (axis, slice) to a dict.
    Dict contains predetermined keys (0, 1, 2). For each key with have a set of int (slice indices to be read)

    volume_shape: shape of the reference volume to perform test on slice indices (if provided)
    """
    indices_to_read_per_axis: dict[int, set[int]] = {
        0: set(),
        1: set(),
        2: set(),
    }
    for slice_ in slices:
        if volume_shape is not None and (slice_.index >= volume_shape[slice_.axis]):
            raise IndexError(
                f"index {slice_.index} is out of bounds for axis {slice_.axis}. (Volume shape is {volume_shape})"
            )
        indices_to_read_per_axis[slice_.axis].add(slice_.index)
    return indices_to_read_per_axis



# coding: utf-8

"""This module is dedicated to instances of :class:`VolumeBase` used at esrf"""

from .edfvolume import EDFVolume  # noqa F401
from .hdf5volume import HDF5Volume  # noqa F401
from .jp2kvolume import JP2KVolume  # noqa F401
from .rawvolume import RawVolume  # noqa F401
from .tiffvolume import MultiTIFFVolume, TIFFVolume  # noqa F401



import json
import numpy
from pyunitsystem.metricsystem import MetricSystem
from tomoscan.utils.drac.mapper import (
    flatten_vol_metadata,
    map_nabu_keys_to_drac,
)


_NABU_VOL_RAW_METADATA = {
    "@NX_class": "NXcollection",
    "date": "2024-10-02T17:42:01",
    "nabu_config": {
        "dataset": {
            "binning": 1,
            "binning_z": 1,
            "darks_flats_dir": "None",
            "exclude_projections": "None",
            "hdf5_entry": "/entry0000",
            "location": "/mnt/multipath-shares/tmp_14_days/payno/test_deployment/bambou_hercules_0001.nx",
            "nexus_version": 1.0,
            "overwrite_metadata": "",
            "projections_subsampling": numpy.array([1, 0]),
        },
        "output": {
            "file_format": "hdf5",
            "file_prefix": "bambou_hercules_0001_vol",
            "float_clip_values": "None",
            "jpeg2000_compression_ratio": "None",
            "location": "/mnt/multipath-shares/tmp_14_days/payno/test_deployment/reconstructed_volumes",
            "overwrite_results": True,
            "tiff_single_file": False,
        },
        "phase": {
            "distance_cm": 210,
            "distance_m": 2.1,
            "ctf_advanced_params": "length_scale=1e-05; lim1=1e-05; lim2=0.2; normalize_by_mean=True",
            "ctf_geometry": "z1_v=None; z1_h=None; detec_pixel_size=None; magnification=True",
            "delta_beta": 100.0,
            "method": "paganin",
            "padding_type": "edge",
            "unsharp_coeff": 0.0,
            "unsharp_method": "gaussian",
            "unsharp_sigma": 0.0,
            "pixel_size_m": 3.0325e-06,
            "pixel_size_microns": 3.0325,
            "energy_kev": 130,
        },
        "pipeline": {
            "resume_from_step": "None",
            "save_steps": "None",
            "steps_file": "None",
            "verbosity": "info",
        },
        "postproc": {"histogram_bins": 1000000, "output_histogram": True},
        "preproc": {
            "autotilt_options": "None",
            "ccd_filter_enabled": False,
            "ccd_filter_threshold": 0.04,
            "detector_distortion_correction": "None",
            "detector_distortion_correction_options": "None",
            "dff_sigma": 0.0,
            "double_flatfield_enabled": False,
            "flat_distortion_correction_enabled": False,
            "flat_distortion_params": "tile_size=100; "
            "interpolation_kind='linear'; "
            "padding_mode='edge'; "
            "correction_spike_threshold=None",
            "flatfield": True,
            "log_max_clip": 10.0,
            "log_min_clip": 1e-06,
            "normalize_srcurrent": True,
            "processes_file": "None",
            "rotate_projections": "None",
            "rotate_projections_center": "None",
            "sino_normalization": "None",
            "sino_normalization_file": "",
            "sino_rings_correction": "None",
            "sino_rings_options": "sigma=1.0 ; levels=10 ; " "padding=False",
            "take_logarithm": True,
            "tilt_correction": "None",
        },
        "reconstruction": {
            "angle_offset": 0.0,
            "angles_file": "None",
            "axis_correction_file": "None",
            "centered_axis": True,
            "clip_outer_circle": True,
            "cor_options": "side='from_file'",
            "cor_slice": "None",
            "enable_halftomo": True,
            "end_x": -1,
            "end_y": -1,
            "end_z": 180,
            "fbp_filter_cutoff": 1.0,
            "fbp_filter_type": "ramlak",
            "iterations": 200,
            "method": "FBP",
            "optim_algorithm": "chambolle-pock",
            "padding_type": "edge",
            "positivity_constraint": True,
            "preconditioning_filter": True,
            "rotation_axis_position": 2060.0,
            "sample_detector_dist": "None",
            "source_sample_dist": "None",
            "start_x": 0,
            "start_y": 0,
            "start_z": 150,
            "translation_movements_file": "None",
            "weight_tv": 0.01,
        },
        "resources": {
            "cpu_workers": 0,
            "gpu_id": numpy.array([], dtype=numpy.float64),
            "gpus": 1,
            "memory_per_node": numpy.array([90.0, 1.0]),
            "method": "local",
            "queue": "gpu",
            "threads_per_node": numpy.array([100.0, 1.0]),
            "walltime": numpy.array([1, 0, 0]),
        },
    },
    "processing_options": {
        "flatfield": {
            "binning": numpy.array([1, 1]),
            "do_flat_distortion": False,
            "flat_distortion_params": {
                "correction_spike_threshold": "None",
                "interpolation_kind": "linear",
                "padding_mode": "edge",
                "tile_size": 100,
            },
            "flats_srcurrent": numpy.array([0.19458202]),
            "normalize_srcurrent": True,
            "projs_indices": numpy.array([81, 82, 83, 4478, 4479, 4480]),
            "radios_srcurrent": numpy.array(
                [0.19458, 0.1945795, 0.19457899, 0.19364, 0.19364, 0.19364],
                dtype=numpy.float32,
            ),
        },
        "histogram": {"histogram_bins": 1000000},
        "read_chunk": {
            "binning": numpy.array([1, 1]),
            "dataset_subsampling": numpy.array([1, 0]),
            "files": {
                "100": "silx:///mnt/multipath-shares/tmp_14_days/payno/test_deployment/bambou_hercules_0001.nx?path=/entry0000/instrument/detector/data&slice=100",
                "999": "silx:///mnt/multipath-shares/tmp_14_days/payno/test_deployment/bambou_hercules_0001.nx?path=/entry0000/instrument/detector/data&slice=999",
            },
            "sub_region": "None",
        },
        "reconstruction": {
            "angles": numpy.array(
                [
                    0.00000000e00,
                    1.43989663e-03,
                    2.85934019e-03,
                    ...,
                    6.27849473e00,
                    6.27991963e00,
                    6.28134453e00,
                ]
            ),
            "axis_correction": "None",
            "centered_axis": True,
            "clip_outer_circle": True,
            "cor_estimated_auto": False,
            "enable_halftomo": True,
            "end_x": 4121,
            "end_y": 4121,
            "end_z": 180,
            "fbp_filter_cutoff": 1.0,
            "fbp_filter_type": "ramlak",
            "method": "FBP",
            "padding_type": "edge",
            "position": numpy.array([0.026, 0.0013, 0.0007]),
            "rotation_axis_position": 2060.0,
            "sample_detector_dist": "None",
            "source_sample_dist": "None",
            "start_x": 0,
            "start_y": 0,
            "start_z": 150,
            "voxel_size_cm": numpy.array([1.25e-05, 2.25e-05, 3.25e-05]),
        },
        "save": {
            "file_format": "hdf5",
            "file_prefix": "bambou_hercules_0001_vol",
            "float_clip_values": "None",
            "jpeg2000_compression_ratio": "None",
            "location": "/mnt/multipath-shares/tmp_14_days/payno/test_deployment/reconstructed_volumes",
            "overwrite": True,
            "overwrite_results": True,
            "tiff_single_file": False,
        },
        "take_log": {"log_max_clip": 10.0, "log_min_clip": 1e-06},
    },
    "reconstruction_stages": {
        "((0, 4400), (150, 181), (0, 2560))": "bambou_hercules_0001_vol/bambou_hercules_0001_vol_00150.hdf5"
    },
}


# here we only compare keys because values have been cast (list to numpy array...) but this shouldn't prevent from creating the nabu config file
def compare_keys(dict_1: dict, dict_2: dict):
    """
    compare two nested dictionaries.
    Only keys are checked.
    Because we want to make sure the structure of the dict are equivalent. Some values might have been casted (from numpy.array to list for example)
    """
    all_keys = set(dict_1.keys()) | set(dict_2.keys())
    for key in all_keys:
        assert key in dict_1
        assert key in dict_2
        if isinstance(dict_1[key], dict):
            assert isinstance(
                dict_2[key], dict
            ), f"fail for key ({key}), values are {type(dict_1[key]), {type(dict_2[key])}}"
            compare_keys(dict_1[key], dict_2[key])


def test_nabu_and_drac_compatibility():
    """
    test that the nabu metadata can be publish to drac/icat and retrieve from drac/icat to nabu config (for remote processing)
    """

    flatten_dict = flatten_vol_metadata(_NABU_VOL_RAW_METADATA)
    drac_dict = map_nabu_keys_to_drac(flatten_dict=flatten_dict)
    from pprint import pprint

    pprint(drac_dict)
    # make sure the flatten dict is serializable
    json.dumps(drac_dict)

    assert drac_dict == {
        "TOMOReconstructionPhase_ctf_advanced_params": "length_scale=1e-05; "
        "lim1=1e-05; lim2=0.2; "
        "normalize_by_mean=True",
        "TOMOReconstructionPhase_ctf_geometry": "z1_v=None; z1_h=None; "
        "detec_pixel_size=None; "
        "magnification=True",
        "TOMOReconstructionPhase_delta_beta": 100.0,
        "TOMOReconstructionPhase_detector_sample_distance": 2100.0,
        "TOMOReconstructionPhase_method": "paganin",
        "TOMOReconstructionPhase_padding_type": "edge",
        "TOMOReconstructionPhase_unsharp_coeff": 0.0,
        "TOMOReconstructionPhase_unsharp_method": "gaussian",
        "TOMOReconstructionPhase_unsharp_sigma": 0.0,
        "TOMOReconstruction_angle_offset": 0.0,
        "TOMOReconstruction_angles_file": "None",
        "TOMOReconstruction_axis_correction_file": "None",
        "TOMOReconstruction_centered_axis": True,
        "TOMOReconstruction_clip_outer_circle": True,
        "TOMOReconstruction_cor_options": "side='from_file'",
        "TOMOReconstruction_enable_halftomo": True,
        "TOMOReconstruction_end_x": -1,
        "TOMOReconstruction_end_y": -1,
        "TOMOReconstruction_end_z": 180,
        "TOMOReconstruction_fbp_filter_cutoff": 1.0,
        "TOMOReconstruction_fbp_filter_type": "ramlak",
        "TOMOReconstruction_method": "FBP",
        "TOMOReconstruction_optim_algorithm": "chambolle-pock",
        "TOMOReconstruction_padding_type": "edge",
        "TOMOReconstruction_preconditioning_filter": True,
        "TOMOReconstruction_rotation_axis_position": 2060.0,
        "TOMOReconstruction_start_x": 0,
        "TOMOReconstruction_start_y": 0,
        "TOMOReconstruction_start_z": 150,
        "TOMOReconstruction_translation_movements_file": "None",
        "TOMOReconstruction_weight_tv": 0.01,
        "endTime": "2024-10-02T17:42:01",
        "TOMOReconstruction_voxel_size_x": 1.25e-05
        * MetricSystem.CENTIMETER.value
        / MetricSystem.MICROMETER.value,
        "TOMOReconstruction_voxel_size_y": 2.25e-05
        * MetricSystem.CENTIMETER.value
        / MetricSystem.MICROMETER.value,
        "TOMOReconstruction_voxel_size_z": 3.25e-05
        * MetricSystem.CENTIMETER.value
        / MetricSystem.MICROMETER.value,
    }



# coding: utf-8


import tomoscan.progress


def test_progress():
    """Simple test of the Progress API"""
    progress = tomoscan.progress.Progress("this is progress")
    progress.reset()
    progress.startProcess()
    progress.setMaxAdvancement(80)
    for adv in (10, 20, 50, 70):
        progress.setAdvancement(adv)
    for _ in range(10):
        progress.increaseAdvancement(1)


def test_advancement():
    """Simple test of the _Advancement API"""
    for i in range(4):
        tomoscan.progress._Advancement.getNextStep(
            tomoscan.progress._Advancement.getStep(i)
        )



"""module with helper ot save single frame"""

from __future__ import annotations

from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase


class SingleFrameSaverHelper:
    """
    Allow to save to a single file per frame structure along any axis. The volume is expected to be 3D.

    .. note:: This is not part of the VolumeSingleFrameBase API because we want to make it clear that way of saving is very particular
              and might comes with some botlnecks
    """

    def __init__(
        self,
        volume: VolumeSingleFrameBase,
        data_shape: tuple,
        dtype,
        cache_size: int | None = None,
    ) -> None:
        """
        :param volume: volume to save data for
        :param shape: final data shape
        :param dtype: data type (as a numpy data type)
        :param cache_size: cache of the size (in bytes). Once this size is reached then the data will be dump to disk...
        """
        if not isinstance(volume, VolumeSingleFrameBase):
            raise TypeError(
                f"volume is expected to be an instance of VolumeSingleFrameBase. Get {type(volume)}"
            )
        if not isinstance(data_shape, tuple):
            raise TypeError(
                f"shape is expected to be an instance of tuple. Get {type(data_shape)}"
            )

        self._volume = volume
        self._shape = data_shape
        self._dtype = dtype
        self._frame_index_to_file = None
        self._initialized = False

    def init_saver(self):
        self._frame_index_to_file = {}
        for i, frame_dumper in enumerate(
            self._volume.data_file_saver_generator(
                n_frames=self._shape[0],
                data_url=self._volume.data_url,
                overwrite=self._volume.overwrite,
            ),
        ):
            self._frame_index_to_file[i] = frame_dumper
        self._initialized = True

        for i in range(len(self._frame_index_to_file) - 1):
            assert self._frame_index_to_file[i] != self._frame_index_to_file[i + 1]

    def __setitem__(self, index, value):
        if not self._initialized:
            raise RuntimeError(
                "the helper should be initialized first. Freezing the shape. Please call 'init_saver' before dumping any data to it"
            )
        if isinstance(index, int):
            # in case we are saving the full frame there is no need for any cache mecanism
            self._frame_index_to_file[index][:] = value
        else:
            raise NotImplementedError



# coding: utf-8

import os
import sys
import tempfile

import h5py
import numpy
import pytest

import tomoscan.validator
from tomoscan.tests.utils import NXtomoMockContext

frame_validators = (
    tomoscan.validator.FlatEntryValidator,
    tomoscan.validator.DarkEntryValidator,
    tomoscan.validator.ProjectionEntryValidator,
)


@pytest.mark.parametrize("validator_cls", frame_validators)
def test_frames_validator(validator_cls):
    """Test frame validator on a complete dataset"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
    ) as scan:
        validator = validator_cls(scan)
        assert validator.is_valid(), "scan contains all kind of frames"


@pytest.mark.parametrize("validator_cls", frame_validators)
def test_frames_validator_2(validator_cls):
    """Test frame validator on a empty dataset"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=0,
        n_ini_proj=0,
        create_ini_dark=False,
        create_ini_ref=False,
        create_final_ref=False,
    ) as scan:
        validator = validator_cls(scan)
        assert not validator.is_valid(), "scan doesn't contains any projection"


@pytest.mark.parametrize("validator_cls", frame_validators)
def test_frames_validator_3(validator_cls):
    """Test frame validator on a dataset missing some projections"""
    tomo_n = 20
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=tomo_n,
        n_ini_proj=tomo_n - 1,
        create_ini_dark=False,
        create_ini_ref=False,
        create_final_ref=False,
    ) as scan:
        with h5py.File(scan.master_file, mode="a") as h5f:
            entry = h5f[scan.entry]
            entry.require_group("instrument").require_group("detector")[
                "tomo_n"
            ] = tomo_n
        validator = validator_cls(scan)
        assert not validator.is_valid(), "scan doesn't contains tomo_n projections"


phase_retrieval_validators = (
    tomoscan.validator.EnergyValidator,
    tomoscan.validator.DistanceValidator,
    tomoscan.validator.PixelValidator,
)


@pytest.mark.parametrize("validator_cls", phase_retrieval_validators)
def test_phase_retrieval_validator(validator_cls):
    """Test dark and flat validator on a complete dataset"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
    ) as scan:
        with h5py.File(scan.master_file, mode="a") as h5f:
            entry_grp = h5f[scan.entry]
            if "instrument/detector/x_pixel_size" in entry_grp:
                del entry_grp["instrument/detector/x_pixel_size"]
            if "instrument/detector/y_pixel_size" in entry_grp:
                del entry_grp["instrument/detector/y_pixel_size"]
        validator = validator_cls(scan)
        assert (
            not validator.is_valid()
        ), "scan have missing energy, distance and pixel size"
        with h5py.File(scan.master_file, mode="a") as h5f:
            beam_grp = h5f[scan.entry].require_group("beam")
            if "incident_energy" in beam_grp:
                del beam_grp["incident_energy"]
            beam_grp["incident_energy"] = 1.0
            beam_grp_2 = h5f[scan.entry].require_group("instrument/beam")
            if "incident_energy" in beam_grp_2:
                del beam_grp_2["incident_energy"]
            beam_grp_2["incident_energy"] = 1.0

            detector_grp = h5f[scan.entry].require_group("instrument/detector")
            if "distance" in detector_grp:
                del detector_grp["distance"]
            detector_grp["distance"] = 1.0

            detector_grp["x_pixel_size"] = 2.0
            detector_grp["y_pixel_size"] = 1.0

        validator.clear()
        assert validator.is_valid(), "scan contains all information for phase retrieval"


frame_values_validators = (
    tomoscan.validator.DarkDatasetValidator,
    tomoscan.validator.FlatDatasetValidator,
    tomoscan.validator.ProjectionDatasetValidator,
)


@pytest.mark.parametrize("validator_cls", frame_values_validators)
def test_frame_broken_vds(validator_cls):
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        create_ini_dark=True,
        create_ini_ref=True,
        create_final_ref=False,
    ) as scan:
        validator = validator_cls(scan=scan, check_vds=True, check_values=False)
        assert (
            validator.is_valid()
        ), "if data is unchanged then validator should valid the entry"
        validator.clear()

        # modify 'data' dataset to set a virtual dataset with broken link (file does not exists)
        with h5py.File(scan.master_file, mode="a") as h5f:
            detector_grp = h5f[scan.entry]["instrument/detector"]
            shape = detector_grp["data"].shape
            del detector_grp["data"]

            # create invalid VDS
            layout = h5py.VirtualLayout(shape=shape, dtype="i4")

            filename = "toto.h5"
            vsource = h5py.VirtualSource(filename, "data", shape=shape)
            layout[0 : shape[0]] = vsource

            detector_grp.create_virtual_dataset("data", layout)

        assert not validator.is_valid(), "should return broken dataset"


@pytest.mark.parametrize("validator_cls", frame_values_validators)
def test_frame_data_with_nan(validator_cls):
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
    ) as scan:
        validator = validator_cls(scan=scan, check_vds=False, check_values=True)
        assert (
            validator.is_valid()
        ), "if data is unchanged then validor should valid the entry"
        validator.clear()

        # modify 'data' dataset to add nan values
        with h5py.File(scan.master_file, mode="a") as h5f:
            data = h5f[scan.entry]["instrument/detector/data"][()]
            del h5f[scan.entry]["instrument/detector/data"]
            data[:] = numpy.nan
            h5f[scan.entry]["instrument/detector/data"] = data

        assert not validator.is_valid(), "should return data contains nan"


high_level_validators = (
    tomoscan.validator.BasicScanValidator,
    tomoscan.validator.ReconstructionValidator,
)


@pytest.mark.parametrize("only_issue", (True, False))
@pytest.mark.parametrize("validator_cls", high_level_validators)
def test_high_level_validators_ok(capsys, validator_cls, only_issue):
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        validator = validator_cls(scan=scan)

        assert validator.is_valid()
        sys.stdout.write(validator.checkup(only_issues=only_issue))
        captured = capsys.readouterr()
        assert "No issue" in captured.out, "check print as been done on stdout"
        validator.clear()


@pytest.mark.parametrize("only_issue", (True, False))
@pytest.mark.parametrize("check_values", (True, False))
@pytest.mark.parametrize("check_dark", (True, False))
@pytest.mark.parametrize("check_flat", (True, False))
@pytest.mark.parametrize("check_phase_retrieval", (True, False))
def test_reconstruction_validator_not_ok(
    capsys, only_issue, check_values, check_dark, check_flat, check_phase_retrieval
):
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
    ) as scan:
        validator = tomoscan.validator.ReconstructionValidator(
            scan=scan,
            check_values=check_values,
            check_flat=check_flat,
            check_dark=check_dark,
            check_phase_retrieval=check_phase_retrieval,
        )

        sys.stdout.write(validator.checkup(only_issues=only_issue))
        captured = capsys.readouterr()
        if check_phase_retrieval:
            assert "2 issues" in captured.out, "should have found 2 issues"
        else:
            "no issue" in captured.out, "should not have found any issue"

        validator.clear()
        # modify 'data' dataset to set nan inside. Now dark, flat and projection should fail
        with h5py.File(scan.master_file, mode="a") as h5f:
            data = h5f[scan.entry]["instrument/detector/data"][()]
            del h5f[scan.entry]["instrument/detector/data"]
            data[:] = numpy.nan
            h5f[scan.entry]["instrument/detector/data"] = data

        sys.stdout.write(validator.checkup(only_issues=only_issue))
        captured = capsys.readouterr()
        n_issues = 0
        if check_phase_retrieval:
            # there is no energy / distance
            n_issues += 2
        if check_values and check_flat:
            # flat contains nan
            n_issues += 1
        if check_values and check_dark:
            # dark contains nan
            n_issues += 1
        if check_values:
            # projections contains nan
            n_issues += 1

        if n_issues == 0:
            "no issue" in captured.out, "should not have found any issue"
        else:
            assert (
                f"{n_issues} issues" in captured.out
            ), f"should have found {n_issues} issues"


def test_validatorbase():
    """Test the Validator base class API"""
    validator = tomoscan.validator.ValidatorBase()

    with pytest.raises(NotImplementedError):
        validator.is_valid()

    with pytest.raises(NotImplementedError):
        validator.run()

    with pytest.raises(NotImplementedError):
        validator.clear()


def test_is_valid_for_reconstruction():
    """test is_valid_for_reconstruction function."""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        assert tomoscan.validator.is_valid_for_reconstruction(
            scan=scan, need_phase_retrieval=True, check_values=True
        ), "This dataset should be valid for reconstruction with phase retrieval"



# coding: utf-8

from tomoscan.esrf.identifier.folderidentifier import (
    BaseFolderAndfilePrefixIdentifierMixIn,
)
from tomoscan.identifier import VolumeIdentifier
from tomoscan.utils import docstring

__all__ = [
    "JP2KVolumeIdentifier",
]


class JP2KVolumeIdentifier(BaseFolderAndfilePrefixIdentifierMixIn, VolumeIdentifier):
    """Identifier specific to JP2K volume"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, tomo_type=VolumeIdentifier.TOMO_TYPE)

    @property
    @docstring(VolumeIdentifier)
    def scheme(self) -> str:
        return "jp2k"

    @staticmethod
    def from_str(identifier):
        from tomoscan.esrf.volume.jp2kvolume import JP2KVolume

        return (
            BaseFolderAndfilePrefixIdentifierMixIn._from_str_to_single_frame_identifier(
                identifier=identifier,
                SingleFrameIdentifierClass=JP2KVolumeIdentifier,
                ObjClass=JP2KVolume,
            )
        )



# coding: utf-8

"""Module dedicated to input / output utils"""

from __future__ import annotations

import logging
import os
from h5py import File as HDF5File  # noqa F401
from silx.io.utils import open as hdf5_open

_logger = logging.getLogger(__name__)

_DEFAULT_SWMR_MODE = None

__all__ = [
    "get_swmr_mode",
    "check_virtual_sources_exist",
]


def get_swmr_mode() -> bool | None:
    """
    Return True if the swmr should be used in the tomoools scope
    """
    swmr_mode = os.environ.get("TOMOTOOLS_SWMR", _DEFAULT_SWMR_MODE)
    if swmr_mode in (None, "None", "NONE"):
        return None
    else:
        return swmr_mode in (
            True,
            "True",
            "true",
            "TRUE",
            "1",
            1,
        )


def check_virtual_sources_exist(fname, data_path):
    """
    Check that a virtual dataset points to actual data.

    :param fname: HDF5 file path
    :param data_path: Path within the HDF5 file

    :return bool res: Whether the virtual dataset points to actual data.
    """
    with hdf5_open(fname) as f:
        if data_path not in f:
            _logger.error(f"No dataset {data_path} in file {fname}")
            return False
        dptr = f[data_path]
        if not dptr.is_virtual:
            return True
        for vsource in dptr.virtual_sources():
            vsource_fname = os.path.join(
                os.path.dirname(dptr.file.filename), vsource.file_name
            )
            if not os.path.isfile(vsource_fname):
                _logger.error(f"No such file: {vsource_fname}")
                return False
            elif not check_virtual_sources_exist(vsource_fname, vsource.dset_name):
                _logger.error(f"Error with virtual source {vsource_fname}")
                return False
    return True



"""Defines mapping rule to be applied for reconstructed volume to drac"""

from __future__ import annotations
import numpy
from pyunitsystem.metricsystem import MetricSystem


class UnitConversion:
    def __init__(self, raw_data_unit, drac_unit) -> None:
        self.raw_data_unit = raw_data_unit
        self.drac_unit = drac_unit

    def to_raw_data(self, value):
        return value * self.drac_unit.value / self.raw_data_unit.value

    def to_drac(self, value):
        return value * self.raw_data_unit.value / self.drac_unit.value


class DracRawDataMapping:
    """
    Map drac/icat metadata from / to raw data. Raw data in this context is expected to be a reconstructed volume (by nabu)
    This layer ease handling of unit conversion as data type (from a single field to multi field structures.)

    A raw data / nabu field can be split into several drac/icat field. And several drac/icat field can be concatenate to a single raw_data/nabu field.
    This is the purpose of the 'raw_data_field_index', 'raw_data_field_size' and 'raw_data_type_constructor' parameters.
    The over way around is not forseen
    """

    def __init__(
        self,
        drac_key: str,
        raw_data_key: str,
        raw_data_field_index: int | None = None,
        raw_data_type_constructor=None,
        raw_data_unit=None,
        drac_unit=None,
    ) -> None:
        self.drac_key = drac_key
        self.raw_data_key = raw_data_key

        self.raw_data_field_index = raw_data_field_index

        self.raw_data_type_constructor = raw_data_type_constructor

        if raw_data_unit is not None:
            assert (
                drac_unit is not None
            ), "both raw_data and drac units should be provided is some conversion is requested"

            self._unit_converter = UnitConversion(
                raw_data_unit=raw_data_unit,
                drac_unit=drac_unit,
            )
        else:
            self._unit_converter = None

    def to_drac(self, raw_value):
        "convert a raw_data (nabu) field to an icat field"
        if self.raw_data_field_index is not None:
            assert isinstance(self.raw_data_field_index, int)
            assert isinstance(raw_value, (list, tuple, numpy.array))
            raw_value = raw_value[self.raw_data_field_index]

        if self._unit_converter is not None:
            return self._unit_converter.to_drac(value=raw_value)
        else:
            return raw_value


DRAC_NABU_MAPPING = (
    DracRawDataMapping(
        drac_key="TOMOReconstruction_angle_offset",  # float
        raw_data_key="nabu_config/reconstruction/angle_offset",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_angles_file",  # str - file path
        raw_data_key="nabu_config/reconstruction/angles_file",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_axis_correction_file",  # str - file path
        raw_data_key="nabu_config/reconstruction/axis_correction_file",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_centered_axis",  # bool
        raw_data_key="nabu_config/reconstruction/centered_axis",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_clip_outer_circle",  # bool
        raw_data_key="nabu_config/reconstruction/clip_outer_circle",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_cor_options",  # str
        raw_data_key="nabu_config/reconstruction/cor_options",
    ),
    DracRawDataMapping(
        drac_key="endTime",  # str
        raw_data_key="date",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_enable_halftomo",  # bool
        raw_data_key="nabu_config/reconstruction/enable_halftomo",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_end_x",  # int (px)
        raw_data_key="nabu_config/reconstruction/end_x",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_end_y",  # int (px)
        raw_data_key="nabu_config/reconstruction/end_y",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_end_z",  # int (px)
        raw_data_key="nabu_config/reconstruction/end_z",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_fbp_filter_cutoff",  # float
        raw_data_key="nabu_config/reconstruction/fbp_filter_cutoff",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_fbp_filter_type",  # str
        raw_data_key="nabu_config/reconstruction/fbp_filter_type",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_method",  # str
        raw_data_key="nabu_config/reconstruction/method",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_optim_algorithm",  # str
        raw_data_key="nabu_config/reconstruction/optim_algorithm",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_padding_type",  # str
        raw_data_key="nabu_config/reconstruction/padding_type",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_preconditioning_filter",  # str
        raw_data_key="nabu_config/reconstruction/preconditioning_filter",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_rotation_axis_position",  # float
        raw_data_key="nabu_config/reconstruction/rotation_axis_position",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_start_x",  # int
        raw_data_key="nabu_config/reconstruction/start_x",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_start_y",  # int
        raw_data_key="nabu_config/reconstruction/start_y",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_start_z",  # int
        raw_data_key="nabu_config/reconstruction/start_z",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_translation_movements_file",  # str
        raw_data_key="nabu_config/reconstruction/translation_movements_file",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_weight_tv",  # float
        raw_data_key="nabu_config/reconstruction/weight_tv",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_voxel_size_x",  # float
        raw_data_key="processing_options/reconstruction/voxel_size_cm",
        raw_data_field_index=0,
        raw_data_unit=MetricSystem.CENTIMETER,
        drac_unit=MetricSystem.MICROMETER,
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_voxel_size_y",  # float
        raw_data_key="processing_options/reconstruction/voxel_size_cm",
        raw_data_field_index=1,
        raw_data_unit=MetricSystem.CENTIMETER,
        drac_unit=MetricSystem.MICROMETER,
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstruction_voxel_size_z",  # float
        raw_data_key="processing_options/reconstruction/voxel_size_cm",
        raw_data_field_index=2,
        raw_data_unit=MetricSystem.CENTIMETER,
        drac_unit=MetricSystem.MICROMETER,
    ),
    # phase
    DracRawDataMapping(
        drac_key="TOMOReconstructionPhase_ctf_advanced_params",  # str
        raw_data_key="nabu_config/phase/ctf_advanced_params",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstructionPhase_ctf_geometry",  # str
        raw_data_key="nabu_config/phase/ctf_geometry",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstructionPhase_delta_beta",  # float
        raw_data_key="nabu_config/phase/delta_beta",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstructionPhase_detector_sample_distance",  # float
        raw_data_key="nabu_config/phase/distance_m",
        raw_data_unit=MetricSystem.METER,
        drac_unit=MetricSystem.MILLIMETER,
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstructionPhase_method",  # str
        raw_data_key="nabu_config/phase/method",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstructionPhase_padding_type",  # ste
        raw_data_key="nabu_config/phase/padding_type",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstructionPhase_unsharp_coeff",  # float
        raw_data_key="nabu_config/phase/unsharp_coeff",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstructionPhase_unsharp_method",  # str
        raw_data_key="nabu_config/phase/unsharp_method",
    ),
    DracRawDataMapping(
        drac_key="TOMOReconstructionPhase_unsharp_sigma",  # float
        raw_data_key="nabu_config/phase/unsharp_sigma",
    ),
)



"""Scan dedicated for bliss format - based on EDF files"""

from __future__ import annotations

import logging
import os
import glob
from typing import Sequence

import numpy

from tomoscan.identifier import ScanIdentifier
from tomoscan.scanbase import TomoScanBase
from tomoscan.utils import docstring


from dataclasses import dataclass, field
from numpy.typing import NDArray, DTypeLike
from silx.io.utils import h5py_read_dataset

try:
    from tqdm.auto import tqdm
except ImportError:
    has_tqdm = False
else:
    has_tqdm = True
import h5py

_logger = logging.getLogger(__name__)

try:
    import imageio as iio
except ImportError as e:
    _logger.warning(
        f"WARNING: error importing Imageio, using matplotlib instead. Error is {e}"
    )
    has_imageio = False
else:
    has_imageio = True

__all__ = [
    "FluoTomoScan",
]


@dataclass
class FluoTomoScan:
    """Dataset manipulation class."""

    scan: str
    dataset_basename: str
    detectors: tuple = ()

    skip_angle_inds: Sequence[int] | NDArray | None = None
    dtype: DTypeLike = numpy.float32
    verbose: bool = False

    angles: NDArray | None = None
    el_lines: dict[str, list[str]] = field(default_factory=dict)
    pixel_size: float | None = None
    energy: float | None = None

    def __post_init__(self):
        self.detected_detectors = tuple(self.detect_detectors())
        self.detect_pixel_size_and_energy()
        _logger.info(f"Detectors: {self.detected_detectors}")
        if len(self.detectors) == 0:
            self.detectors = self.detected_detectors
        for det in self.detectors:
            if det not in self.detected_detectors:
                raise ValueError(
                    f"The detector should be in {self.detected_detectors} and is {self.detectors}"
                )
        self.detect_elements()

        if self.skip_angle_inds is not None:
            self.skip_angle_inds = numpy.array(self.skip_angle_inds)

        if self.angles is None:
            self.detect_rot_angles()

    @property
    def rot_angles_deg(self) -> NDArray:
        if self.angles is None:
            raise ValueError("Rotation angles not initialized")

        if self.skip_angle_inds is None:
            return self.angles
        else:
            return numpy.delete(self.angles, self.skip_angle_inds)

    @property
    def rot_angles_rad(self) -> NDArray:
        return numpy.deg2rad(self.rot_angles_deg)

    def detect_rot_angles(self):
        tmp_angles = []
        proj_ind = 0
        while True:
            proj_ind += 1
            prj_dir = os.path.join(
                self.scan, "fluofit", self.dataset_basename + "_%03d" % proj_ind
            )
            info_file = os.path.join(prj_dir, "info.txt")
            if not os.path.exists(info_file):
                _logger.debug(
                    f"{info_file} doesn't exist, while expected to be present in each projection folder."
                )
                break
            with open(info_file, "r") as f:
                info_str = f.read()
            tmp_angles.append(float(info_str.split(" ")[2]))
        _logger.info(
            f"Found angle information in info file for {proj_ind} projections."
        )
        self.angles = numpy.array(tmp_angles, ndmin=1, dtype=numpy.float32)

    def detect_pixel_size_and_energy(self):
        pixel_size_path = os.path.join(self.scan, self.dataset_basename + "_%03d" % 1)
        h5_files = glob.glob1(pixel_size_path, "*.h5")
        if len(h5_files) > 1:
            raise ValueError(
                "More than one hdf5 file in scan directory. Expect only ONE to pick pixel size."
            )
        elif len(h5_files) == 0:
            pattern = os.path.join(pixel_size_path, "*.h5")
            raise ValueError(
                f"Unable to find the hdf5 file in scan directory to pick pixel size. RegExp used is {pattern}"
            )
        else:
            try:
                if "3DXRF" in h5_files[0]:
                    sample_name = h5_files[0].split("_3DXRF_")[0].split("-")[1]
                else:
                    sample_name = h5_files[0].split("_XRF_")[0].split("-")[1]
            except IndexError:
                raise ValueError(
                    f"unable to deduce sample name from {h5_files[0]}. Expected synthax is 'proposal-samplename_XRF_XXX.h5'"
                )
            entry_name = (
                "entry_0000: "
                + sample_name
                + " - "
                + self.dataset_basename
                + "_%03d" % 1
            )
        with h5py.File(os.path.join(pixel_size_path, h5_files[0]), "r") as f:
            self.pixel_size = (
                h5py_read_dataset(f[entry_name]["FLUO"]["pixelSize"]) * 1e-6
            )
            self.energy = float(
                h5py_read_dataset(
                    f[entry_name]["instrument"]["monochromator"]["energy"]
                )
            )

    def detect_detectors(self):
        proj_1_dir = os.path.join(
            self.scan, "fluofit", self.dataset_basename + "_%03d" % 1
        )
        detected_detectors = []
        file_names = glob.glob1(proj_1_dir, "IMG_*area_density_ngmm2.tif")
        for file in file_names:
            det = file.split("_")[1]
            if det not in detected_detectors:
                detected_detectors.append(det)
        return detected_detectors

    def detect_elements(self):
        proj_1_dir = os.path.join(
            self.scan, "fluofit", self.dataset_basename + "_%03d" % 1
        )
        detector = self.detectors[0]
        file_names = glob.glob1(proj_1_dir, f"IMG_{detector}*area_density_ngmm2.tif")
        for file in file_names:
            el_str = file.split("_")[2]
            element, line = el_str.split("-")
            try:
                if line not in self.el_lines[element]:
                    self.el_lines[element].append(line)
                    self.el_lines[element] = sorted(self.el_lines[element])
            except KeyError:
                self.el_lines[element] = [line]

    def load_data(self, det: str, element: str, line_ind: int = 0):
        if not has_imageio:
            raise RuntimeError("imageio not install. Cannot load data.")
        if det not in self.detectors:
            raise RuntimeError(
                f"The detector {det} is invalid. Valid ones are {self.detectors}"
            )
        if self.angles is None:
            raise RuntimeError("Rotation angles not initilized")

        if self.detectors is None:
            raise RuntimeError("Detectors not initialized")

        line = self.el_lines[element][line_ind]

        data_det = []

        description = f"Loading images of {element}-{line} ({det}): "

        if has_tqdm:
            angles_iterator = tqdm(
                range(len(self.angles)), disable=self.verbose, desc=description
            )
        else:
            angles_iterator = range(len(self.angles))

        for ii_i in angles_iterator:
            if self.skip_angle_inds is not None and ii_i in self.skip_angle_inds:
                if self.verbose:
                    _logger.info(f"Skipping angle index: {ii_i}")
                continue

            proj_dir = os.path.join(
                self.scan, "fluofit", self.dataset_basename + "_%03d" % (ii_i + 1)
            )
            img_path = os.path.join(
                proj_dir, f"IMG_{det}_{element}-{line}_area_density_ngmm2.tif"
            )

            if self.verbose:
                _logger.info(f"Loading {ii_i+1}/{len(self.angles)}: {img_path}")

            img = iio.imread(img_path)
            data_det.append(numpy.nan_to_num(numpy.array(img, dtype=self.dtype)))

        data = numpy.array(data_det)
        return numpy.ascontiguousarray(data)

    @staticmethod
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        raise NotImplementedError("Not implemented for fluo-tomo yet.")

    @docstring(TomoScanBase)
    def get_identifier(self) -> ScanIdentifier:
        raise NotImplementedError("Not implemented for fluo-tomo yet.")



from tomoscan.utils.io import deprecated_warning

deprecated_warning(
    type_="Module",
    name="tomoscan.nexus.paths.nxinstrument",
    reason="dedicated project created",
    replacement="nxtomo.paths.nxinstrument",
    since_version=2.0,
)

from nxtomo.paths.nxinstrument import *  # noqa F401



import functools


def _docstring(dest, origin):
    """Implementation of docstring decorator.

    It patches dest.__doc__.
    """
    if not isinstance(dest, type) and isinstance(origin, type):
        # func is not a class, but origin is, get the method with the same name
        try:
            origin = getattr(origin, dest.__name__)
        except AttributeError:
            raise ValueError(f"origin class has no {dest.__name__} method")

    dest.__doc__ = origin.__doc__
    return dest


def docstring(origin):
    """Decorator to initialize the docstring from another source.

    This is useful to duplicate a docstring for inheritance and composition.

    If origin is a method or a function, it copies its docstring.
    If origin is a class, the docstring is copied from the method
    of that class which has the same name as the method/function
    being decorated.

    :param origin:
        The method, function or class from which to get the docstring
    :raises ValueError:
        If the origin class has not method n case the
    """
    return functools.partial(_docstring, origin=origin)
